Require Import Metatheory.

Require Import Arith.
Require Import Decidable.
Require Import List.

Require Import actions.
Require Import indexes.
Require Import nat_facts.

Open Scope nat_scope.

Set Implicit Arguments.

Hint Constructors index.

Section Events.
  Variable val : Set.

  (* event tr t i j holds if the ith action for tid t in erased trace tr occurs at
     index j in trace. *)
  Inductive event : erased_trace val -> tid -> nat -> nat -> Prop :=
  | event_0 : forall tr t a,
    event ((t, a) :: tr) t 0 0
  | event_eq_S : forall tr t i j a,
    event tr t i j ->
    event ((t, a) :: tr) t (S i) (S j)
  | event_neq_S : forall tr t t' i j a,
    event tr t i j ->
    t <> t' ->
    event ((t', a) :: tr) t i (S j).

  Hint Constructors event.

  (* An event index is valid if there is a corresponding trace index for the event
     trace event index. *)
  Definition event_valid tr t i :=
    exists j, event tr t i j.

  (* The event index is always less than the trace length. *)
  Lemma event_lt_length : forall tr t i j,
    event tr t i j ->
    j < length tr.
  Proof.
    intros tr t i j evt.
    induction evt; simpl; auto with arith.
  Qed.

  (* There is some event index for every action in the trace. *)
  Lemma lt_length_event : forall tr j,
    j < length tr ->
    exists t, exists i, event tr t i j.
  Proof.
    intros tr j length_tr.

    generalize dependent j.
    induction tr as [ | [ t a ] tr ]; intros j j_lt.

    simpl in j_lt. inversion j_lt.

    induction j as [ | j ].
    eauto.
    simpl in j_lt. apply lt_S_n in j_lt.
    pose proof IHtr j j_lt as [ t' [ i evt ] ].
    exists t'.
    destruct (eq_var_dec t t'); subst; eauto.
  Qed.

  Lemma event_index : forall tr t i j,
    event tr t i j ->
    exists a, index tr j (t, a).
  Proof.
    intros tr t i j evt.
    induction evt; eauto; destruct IHevt; eauto.
  Qed.

  Lemma index_event : forall tr j t a,
    index tr j (t, a) ->
    exists i, event tr t i j.
  Proof.
    intros tr j t a idx.
    pose proof index_lt_length idx as lt_length.
    pose proof lt_length_event _ lt_length as [ t' [ i evt ] ].
    pose proof event_index evt as [ a' idx' ].
    assert ((t, a) = (t', a')) as eq.
    eapply index_unique; eauto.
    inversion eq. subst. eauto.
  Qed.

  (* Each event index corresponds to exactly one index. *)
  Lemma event_unique_index : forall tr t i j j',
    event tr t i j ->
    event tr t i j' ->
    j = j'.
  Proof.
    intros tr t i j j' evt evt'.
    generalize dependent j'.
    induction evt; intros j' evt'; subst;
      inversion evt'; subst; auto; tauto.
  Qed.

  (* Each index corresponds to exactly one event index. *)
  Lemma index_unique_event :
    forall tr t1 i1 t2 i2 j,
      event tr t1 i1 j ->
      event tr t2 i2 j ->
      t1 = t2 /\ i1 = i2.
  Proof.
    intros tr t1 i1 t2 i2 j evt evt2.

    generalize dependent j.
    generalize dependent i2.
    generalize dependent t2.
    generalize dependent i1.
    generalize dependent t1.
    induction tr as [ | [ t a ] tr ]; intros t1 i1 t2 i2 j evt1 evt2.

    (* Base case: tr = nil *)
    inversion evt1.

    (* Inductive case: j = S j *)
    inversion evt1 as [ | ? ? ? ? ? evt1' | ? ? ? ? ? ? evt1' ]; subst;
      inversion evt2 as [ | ? ? ? ? ? evt2' | ? ? ? ? ? ? evt2' ]; subst;
        auto;
          pose proof IHtr _ _ _ _ _ evt1' evt2' as [ ? ? ];
            subst; auto; tauto.
  Qed.

  (* An event index (t, i) is valid for tr if and only if there are at least
     i - 1 actions for thread i in tr. *)
  Lemma valid_iff_lt_length_projection : forall tr t i,
    event_valid tr t i <->
    i < length (trace_projection tr t).
  Proof with eauto.
    intros tr t i. split.

    (* If valid, then i < length (projection tr t). *)
    intros [ j evt ].
    induction evt as [ | | ? ? t' ]; simpl;
      [ destruct (eq_var_dec t t)
        | destruct (eq_var_dec t t)
        | destruct (eq_var_dec t t') ];
      simpl; auto with arith; tauto.

    (* If i < length (projection tr t), then valid. *)
    intros i_lt.
    generalize dependent i.
    induction tr as [ | [ t' a ] tr ]; intros i i_lt.

    (* Base case: tr = nil *)
    simpl in i_lt. inversion i_lt.

    (* Inductive case : tr = a :: tr *)
    unfold event_valid in *. simpl in *.
    destruct (eq_var_dec t t'); subst; simpl in *.

    (* The first action in the trace is by thread t. *)
    induction i as [ | i ]...
    pose proof IHtr _ (lt_S_n _ _ i_lt) as [ j ]...

    (* The first action in the trace is not by threas t. *)
    pose proof IHtr _ i_lt as [ j ]...
  Qed.

  (* Given an event index, there exists an action that the event index corresponds
     to, and a trace index for the projection. *)
  Lemma event_inversion : forall tr t i j,
    event tr t i j ->
    exists a, index tr j (t, a) /\ index (trace_projection tr t) i a.
  Proof.
    intros tr t i j evt.
    induction evt as [ ? ? a | ? ? i j a | ? ? t' ? ? a ]; simpl.

    (* Base case : i = 0 *)
    destruct (eq_var_dec t t); eauto; tauto.

    (* Inductive case #1: the first action in tr is by thread t *)
    destruct IHevt as [ a' [ idx idx_proj ] ].
    destruct (eq_var_dec t t); eauto; tauto.

    (* Inductive case #2: the first action in tr is not by thread t *)
    destruct IHevt as [ a' [ idx idx_proj ] ].
    destruct (eq_var_dec t t'); eauto; tauto.
  Qed.

  (* A helpful lemma for dealing with event indexes without having to use
     induction. *)
  Lemma event_length : forall tr1 a tr2 t,
    event (tr1 ++ ((t, a) :: nil) ++ tr2) t (length (trace_projection tr1 t))
    (length tr1).
  Proof with auto.
    intros tr1 a tr2 t.

    (* Proof by induction on tr1. *)
    induction tr1 as [ | [ t' a' ] tr1' ]; simpl...

    (* tr1 = a' :: tr1' *)
    destruct (eq_var_dec t t'); subst; simpl...
  Qed.

  (* Inverting a trace event index for an append, in the case where the index
     falls in the first half of the trace. *)
  Lemma event_append_inversion_1 : forall tr1 tr2 t i j,
    event (tr1 ++ tr2) t i j ->
    j < length tr1 ->
    event tr1 t i j.
  Proof.
    intros tr1 tr2 t i j.

    (* Proof by induction on tr1. *)
    generalize dependent j.
    generalize dependent i.
    induction tr1; intros i j evt j_lt.

    (* tr1 = nil *)
    simpl in *. inversion j_lt.

    (* tr1 = a :: tr1' *)
    inversion evt; subst; eauto with arith.
  Qed.

  (* Inverting a trace event index for an append, in the case where the index
     falls in the second half of the trace. *)
  Lemma event_append_inversion_2 : forall tr1 tr2 t i j,
    event (tr1 ++ tr2) t i (j + length tr1) ->
    exists i', i = i' + length (trace_projection tr1 t) /\
      event tr2 t i' j.
  Proof.
    intros tr1 tr2 t.

    (* Proof by induction on the length of tr1. *)
    induction tr1 as [ | [ t' a ] tr1' ]; intros i j evt.

    (* tr1 = nil *)
    simpl in *. rewrite plus_0_r in evt. eauto.

    (* tr1 = a :: tr1' *)
    inversion evt as [ ? ? ? [] [] [] false | ? ? i' j' ? evt' [] [] i'_eq j'_eq
      | ? ? ? i' j' ? evt' ? [] [] i'_eq j'_eq ].

    (* j = 0 - vacuous *)
    rewrite plus_comm in false. inversion false.

    (* The first element of tr1 is by thread t. *)
    rewrite S_eq_plus_S in j'_eq. subst j'.
    pose proof IHtr1' i' j  evt' as [ i'' [ i''_eq evt'' ] ].
    simpl. destruct (eq_var_dec t t). simpl.
    esplit. split. rewrite S_eq_plus_S. eauto. auto. tauto.

    (* The first element of thread 1 is not by thread t. *)
    rewrite S_eq_plus_S in j'_eq. subst i' j'.
    pose proof IHtr1' i j  evt' as [ i' [ i'_eq evt'' ] ].
    simpl. destruct (eq_var_dec t t'). tauto. eauto.
  Qed.

  (* This lemma is useful for when we have a trace event index for a trace
     constructed using append.  It says that the trace index is either in the
     first half or the second half of the trace. *)
  Lemma event_append_inversion : forall tr1 tr2 t i j,
    event (tr1 ++ tr2) t i j ->
    (j < length tr1 /\ event tr1 t i j) \/
    (exists i', exists j', j = j' + length tr1 /\
      i = i' + length (trace_projection tr1 t) /\
      event tr2 t i' j').
  Proof.
    intros tr1 tr2 t i j evt.
    destruct (lt_or_plus_dec j (length tr1)) as [ | [ j' j'_eq ] ].
    eauto using event_append_inversion_1.
    subst j.
    pose proof event_append_inversion_2 _ _ _ evt as [ i' [ i'_eq evt' ] ].
    subst i.
    eauto 6.
  Qed.

  (* Inverting a trace event index for an append, in the case where the index
     falls in the first half of the trace. *)
  Lemma event_append_1 : forall tr1 tr2 t i j,
    event tr1 t i j ->
    event (tr1 ++ tr2) t i j.
  Proof.
    intros tr1 tr2 t i j evt.
    induction evt; simpl; auto.
  Qed.

  Lemma event_append_2 : forall tr1 tr2 t i j,
    event tr2 t i j ->
    event (tr1 ++ tr2) t (i + length (trace_projection tr1 t))
    (j + length tr1).
  Proof with eauto.
    intros tr1 tr2 t.

    (* Proof by induction on the length of tr1. *)
    induction tr1 as [ | [ t' a ] tr1' ]; intros i j evt.

    (* tr1 = nil *)
    simpl in *. repeat rewrite plus_0_r...

    (* tr1 = a :: tr1' *)
    pose proof IHtr1' _ _ evt.
    simpl.
    destruct (eq_var_dec t t'); subst; simpl;
      repeat rewrite <- plus_n_Sm...
  Qed.

  (* Trace event index validity is decidable. *)
  Lemma event_valid_dec : forall tr t i,
    decidable (event_valid tr t i).
  Proof with auto.
    intros tr t i.
    destruct (le_gt_dec (length (trace_projection tr t)) i) as [ i_ge | i_lt ].

    right. apply le_not_lt in i_ge.
    rewrite <- valid_iff_lt_length_projection in i_ge...

    left. rewrite valid_iff_lt_length_projection...
  Qed.
End Events.
