(** This file defines a relation between write-buffering heaps and sequentially-
   consistent heaps.  Specifically, H_wb "reduces to" H_sc if you can reach H_sc
   by committing all of the buffered values in H_wb.  One write-buffering heap
   may be related to many sequentially-consistent heaps, but we sidestep this
   problem by defining the reduction relation only for coherent heaps. For
   coherent heaps, the heap reduction relation is one-to-one.
   *)

Require Import coherent.
Require Import init_wb.
Require Import heap_relation.
Require Import wb_type_preservation.

Set Implicit Arguments.

Module WBPres := WB_Type_Preservation.

(** We define an equality between weak and strong stores.  We say two stores are
   equal if they have the same domain and the same value for each location in
   the domain. *)
Definition store_eq (S_wb : WB.Store.store) (S_sc : SC.Store.store) : Prop :=
  (forall r, (WB.Store.fresh S_wb r <-> SC.Store.fresh S_sc r))
  /\ (forall r v, WB.Store.get S_wb r v <-> SC.Store.get S_sc r v).

(* The empty stores are equal. *)
Lemma store_eq_empty : store_eq WB.Store.empty SC.Store.empty.
Proof.
  split.
  intro. split; intro; auto with stores.
  intros. split; intros false.
  contradiction (WB.Store.not_fresh_after_get false). auto with stores.
  contradiction (SC.Store.not_fresh_after_get false). auto with stores.
Qed.

(** A weak heap and a strong heap are equal if their stores are equal, the
   weak heap's buffer is empty, and their lock maps are equal. *)
Definition heap_eq (H_wb : WB.heap) (H_sc : SC.heap) : Prop :=
  store_eq (fst (fst H_wb)) (fst H_sc)
  /\ WB.Buffer.is_empty (snd (fst H_wb))
  /\ snd H_wb = snd H_sc.

(** The main relation between write-buffering heaps and sequentially-consistent
   heaps. *)
Definition heap_reduce (H_wb : WB.heap) (H_sc : SC.heap) : Prop :=
  exists H_wb', heap_star H_wb H_wb'
    /\ heap_eq H_wb' H_sc.

(* The heap-reduce relation is only defined for coherent weak heaps. *)
Lemma heap_reduce_implies_coherent : forall H_wb H_sc,
  heap_reduce H_wb H_sc ->
  coherent_heap H_wb.
Proof.
  intros H_wb H_sc reduce.
  destruct reduce as [ H_wb' [ star eq ] ].
  unfold heap_star in star. inversion star.

  subst. destruct eq as [ stores_eq [ buffer_empty lock_maps_eq ] ].
  unfold WB.Buffer.is_empty in buffer_empty.
  unfold coherent_heap, empty_except. auto.

  eapply proj1. apply heap_relation_implies_coherent. eauto.
Qed.

(* If H_wb reduces to H_sc, then their lock maps must be equal. *)
Lemma heap_reduce_implies_lock_maps_equal :
  forall S_wb B_wb Lambda_wb S_sc Lambda_sc,
    heap_reduce (S_wb, B_wb, Lambda_wb) (S_sc, Lambda_sc) ->
    Lambda_wb = Lambda_sc.
Proof.
  intros S_wb B_wb Lambda_wb S_sc Lambda_sc reduce.
  destruct reduce as [ H_wb' [ star eq ] ].
  destruct eq as [ stores_eq [ buffers_eq ? ] ].
  simpl in *. subst Lambda_sc.
  remember (S_wb, B_wb, Lambda_wb) as H_wb.
  rewrite <- heap_relation_star_implies_lock_maps_equal with (H1 := H_wb).
  subst. simpl. auto. auto.
Qed.

(* Changing the lock map does not affect heap reduction. *)
Lemma heap_reduce_independent_of_lock_map :
  forall S_wb B_wb Lambda S_sc Lambda',
    heap_reduce (S_wb, B_wb, Lambda) (S_sc, Lambda) ->
    heap_reduce (S_wb, B_wb, Lambda') (S_sc, Lambda').
Proof.
  intros S_wb B_wb Lambda S_sc Lambda' reduce.
  destruct reduce as [ H_wb' [ star heaps_eq ] ].
  destruct H_wb' as [ [ S_wb' B_wb' ] Lambda'' ].
  pose proof heap_relation_star_implies_lock_maps_equal star as eq.
  simpl in eq. subst Lambda''.
  exists (S_wb', B_wb', Lambda').
  split. eapply heap_relation_star_independent_of_lock_map. eauto.
  destruct heaps_eq as [ ? [ ? ? ] ].
  unfold heap_eq. auto.
Qed.

(* The empty heap reduces to the empty heap. *)
Lemma heap_reduce_empty :
  heap_reduce WB.empty_heap SC.empty_heap.
Proof.
  exists WB.empty_heap.
  split. apply Rstar_0.
  split. apply store_eq_empty.
  simpl. split. apply WB.Buffer.empty_is_empty.
  auto.
Qed.

(* Updating the store in the weak and store heaps preserves heap equality. *)
Lemma update_preserves_heap_eq : forall S_wb B Lambda_wb S_sc Lambda_sc r v,
  heap_eq (S_wb, B, Lambda_wb) (S_sc, Lambda_sc) ->
  heap_eq (WB.Store.update S_wb r v, B, Lambda_wb)
  (SC.Store.update S_sc r v, Lambda_sc).
Proof.
  intros S_wb B Lambda_wb S_sc Lambda_sc r v [ stores_eq [ ? ? ] ].
  unfold heap_eq, store_eq in *.
  destruct stores_eq as [ fresh_iff get_iff ]. simpl in *.
  split. split.
  intros r'. pose proof fresh_iff r' as [ ? ? ].
  destruct (eq_var_dec r r') as [ | r_neq_r' ].
  subst r'. split; intro false.
  contradiction (WB.Store.not_fresh_after_update false).
  contradiction (SC.Store.not_fresh_after_update false).
  apply sym_not_eq in r_neq_r'. split; eauto with stores.
  intros r' v'. pose proof get_iff r' v' as [ ? ? ].
  destruct (eq_var_dec r r') as [ | r_neq_r' ].
  subst r'. split; intro.
  assert (v = v'). eapply WB.Store.get_unique.
  apply WB.Store.get_update_same. eauto. subst v'. eauto with stores.
  assert (v = v'). eapply SC.Store.get_unique.
  apply SC.Store.get_update_same. eauto. subst v'. eauto with stores.
  split; eauto with stores.
  auto.
Qed.

(* An alloc step preserves heap equality. *)
Lemma alloc_preserves_heap_eq : forall H_wb1 H_wb2 H_sc1 H_sc2 t r v,
  heap_eq H_wb1 H_sc1 ->
  WB.heap_step H_wb1 (Some (t, alloc r v)) H_wb2 ->
  SC.heap_step H_sc1 (Some (t, alloc r v)) H_sc2 ->
  heap_eq H_wb2 H_sc2.
Proof.
  intros H_wb1 H_wb2 H_sc1 H_sc2 t r v eq1 steps_wb steps_sc.
  inversion steps_wb; subst.
  inversion steps_sc; subst.
  apply update_preserves_heap_eq. auto.
Qed.

(* An alloc step preserves heap reduction. *)
Lemma alloc_preserves_heap_reduce : forall Sigma H_wb H_wb' H_sc H_sc' t r v,
  WBPres.well_typed Sigma H_wb ->
  heap_reduce H_wb H_sc ->
  WB.step H_wb (Some (t, alloc r v)) H_wb' ->
  SC.step H_sc (Some (t, alloc r v)) H_sc' ->
  heap_reduce H_wb' H_sc'.
Proof.
  intros Sigma H_wb1 H_wb1' H_sc1 H_sc1' t r v typ1 reduce1 steps_WB steps_SC.
  unfold heap_reduce, heap_star in *.
  destruct reduce1 as [ H_wb2 [ reln12 eq21 ] ].
  pose proof alloc_heap_relation_star_diamond Sigma typ1 reln12 steps_WB
    as [ H_wb2' [ reln12' steps_WB2 ] ].
  exists H_wb2'. split. auto. eapply alloc_preserves_heap_eq; eauto.
Qed.

(* A write step preserves heap reduction. *)
Lemma write_preserves_heap_reduce : forall H_wb H_wb' H_sc H_sc' t r v,
  heap_reduce H_wb H_sc ->
  empty_except (snd (fst H_wb)) t r ->
  WB.heap_step H_wb (Some (t, write r v)) H_wb' ->
  SC.heap_step H_sc (Some (t, write r v)) H_sc' ->
  heap_reduce H_wb' H_sc'.
Proof.
  intros H_wb H_wb' H_sc H_sc' t r v reduce except steps_wb steps_sc.
  pose proof heap_reduce_implies_coherent reduce as coherent.
  pose proof write_preserves_coherent coherent except steps_wb as coherent'.
  unfold heap_reduce, heap_star in *.
  destruct reduce as [ H_wb2 [ reln heaps_eq ] ].
  pose proof write_heap_relation_star_diamond except reln
    steps_wb as [ H_wb2' [ reln2' steps_wb2 ] ].
  pose proof heap_relation_star_preserves_coherent coherent' reln2'
    as coherent2'.
  destruct H_wb2 as [ [ S_wb2 B_wb2 ] Lambda_wb2 ].
  destruct heaps_eq as [ ? [ empty_B_wb2 ? ] ].
  inversion steps_wb2; subst.
  pose proof WB.Buffer.dequeue_enqueue_empty t r v empty_B_wb2
    as [ B_wb3 [ B_wb3_eq empty_B_wb3 ] ].
  exists (WB.Store.update S_wb2 r v, B_wb3, Lambda_wb2).
  split.
  pose proof Rstar_equiv_Rstar1 _ heap_relation as [ contains1 contains2 ].
  unfold contains in *.
  apply contains1. eapply Rstar1_n. apply contains2. apply reln2'.
  apply Rstar1_1. apply heap_relation_step.
  eapply heap_relation_star_preserves_coherent. eauto.
  apply Rstar_0.
  apply WB.step_commit with (t := t). auto.
  inversion steps_sc; subst.
  apply update_preserves_heap_eq.
  unfold heap_eq. simpl in *. auto.
Qed.

(* An empty step preserves heap reduction. *)
Lemma empty_step_preserves_heap_reduce : forall H_wb H_wb' H_sc,
  heap_reduce H_wb H_sc ->
  WB.heap_step H_wb None H_wb' ->
  heap_reduce H_wb' H_sc.
Proof.
  intros H_wb H_wb' H_sc reduce steps.
  pose proof heap_reduce_implies_coherent reduce as coher.
  destruct reduce as [ H_wb'' [ star heaps_eq ] ].
  unfold heap_reduce.
  exists H_wb''. split; auto.
  pose proof heap_relation_step coher steps as reln.
  assert (coherent _ heap_relation H_wb' H_wb'') as [ H_wb''' [ ? ? ] ].
  pose proof heap_relation_confluent as confl.
  unfold Confluent, confluent in confl.
  apply confl with (x := H_wb).
  apply Rstar_n with (y := H_wb').
  auto. apply Rstar_0. auto.
  assert (H_wb'' = H_wb''').
  eapply empty_buffer_heap_relation_star.
  destruct heaps_eq as [ ? [ ? ? ] ]. auto. auto. subst H_wb'''.
  auto.
Qed.

(* If r is fresh for H_wb, and H_wb reduces to H_sc, then r is fresh for H_sc.
   *)
Lemma heap_reduce_implies_fresh :
  forall Sigma S_wb B_wb Lambda_wb S_sc Lambda_sc r,
    WBPres.well_typed Sigma (S_wb, B_wb, Lambda_wb) ->
    heap_reduce (S_wb, B_wb, Lambda_wb) (S_sc, Lambda_sc) ->
    WB.Store.fresh S_wb r ->
    SC.Store.fresh S_sc r.
Proof.
  intros Sigma S_wb B_wb Lambda_wb S_sc Lambda_sc r typ reduce fresh_r.
  destruct reduce as [ [ [ S_wb' B_wb' ] Lambda_wb' ] [ star eq ] ].
  pose proof heap_relation_star_preserves_fresh _ typ fresh_r star
    as fresh_r'. simpl in fresh_r'.
  destruct eq as [ [ fresh_iff ? ] [ ? ? ] ].
  destruct (fresh_iff r) as [ ? ? ].
  simpl in *. auto.
Qed.

(* If r is in the domain of S_wb, then it is also in the domain of S_sc. *)
Lemma heap_reduce_implies_get_exists :
  forall S_wb B_wb Lambda_wb S_sc Lambda_sc r v,
    heap_reduce (S_wb, B_wb, Lambda_wb) (S_sc, Lambda_sc) ->
    WB.Store.get S_wb r v ->
    exists v', SC.Store.get S_sc r v'.
Proof.
  intros S_wb B_wb Lambda_wb S_sc Lambda_sc r v reduce get_r.
  destruct reduce as [ [ [ S_wb' B_wb' ] Lambda_wb' ] [ star eq ] ].
  pose proof heap_relation_star_preserves_get_exists star get_r
    as [ v' get_r' ].
  simpl in get_r'.
  destruct eq as [ [ ? get_iff ] [ ? ? ] ].
  rewrite get_iff in get_r'. simpl in get_r'. eauto.
Qed.

(* If r maps to v in H_wb's store, and H_wb has no values buffered for r, and
   H_wb reduces to H_sc, then r maps to v in H_sc's store. *)
Lemma empty_buffer_heap_reduce_implies_get :
  forall S_wb B_wb Lambda_wb S_sc Lambda_sc r v,
    heap_reduce (S_wb, B_wb, Lambda_wb) (S_sc, Lambda_sc) ->
    WB.Buffer.is_empty_loc B_wb r ->
    WB.Store.get S_wb r v ->
    SC.Store.get S_sc r v.
Proof.
  intros S_wb B_wb Lambda_wb S_sc Lambda_sc r v reduce empty_B get_wb.
  destruct reduce as [ [ [ S_wb' B_wb' ] Lambda_wb' ] [ star eq ] ].
  assert (WB.Store.get (fst (fst (S_wb, B_wb, Lambda_wb))) r v) as get_wb2.
  simpl. auto.
  pose proof heap_relation_star_preserves_get get_wb2 empty_B star as get_wb'.
  destruct eq as [ [ ? get_iff ] ].
  pose proof get_iff r v as [ ? ? ].
  simpl in *. auto.
Qed.

(* If r maps to v in H_wb's store, and H_wb has no values buffered for r, and
   H_wb reduces to H_sc, then r maps to v in H_sc's store. *)
Lemma top_heap_reduce_implies_get :
  forall S_wb B_wb Lambda_wb S_sc Lambda_sc t r v,
    heap_reduce (S_wb, B_wb, Lambda_wb) (S_sc, Lambda_sc) ->
    WB.Buffer.top B_wb t r = Some v ->
    SC.Store.get S_sc r v.
Proof with eauto.
  intros S_wb B_wb Lambda_wb S_sc Lambda_sc t r v reduce top_B.
  forwards coher : heap_reduce_implies_coherent...
  forwards not_empty : WB.Buffer.top_not_empty...
  unfold coherent_heap in coher.
  specializes coher (>>> __ __ __).
  simpl... simpl in coher.
  unfold empty_except in coher.
  destruct reduce as [ [ [ S_wb' B_wb' ] Lambda_wb' ] [ star eq ] ].
  destruct eq as [ eq_S [ empty_B ? ] ].
  simpl in *.
  forwards top' : heap_relation_star_preserves_top_value...
  eapply top_value_top...
  inversion top' as [ ? ? ? ? ? ? ? top'' | ]; subst.
  apply WB.Buffer.top_not_empty in top''.
  simpl in empty_B.
  contradiction top''...
  unfold store_eq in eq_S.
  destruct eq_S as [ ? eq_read ].
  rewrite <- eq_read...
Qed.

