Require Import coherent.
Require Import init_wb.
Require Import wb_type_preservation.

Set Implicit Arguments.

Open Scope nat_scope.

Import WB.
Import WB_Type_Preservation.

(* We define a relation over heaps. A heap H1 is related to H2 if you can get to
   H2 by taking a single empty step. The relation is defined only for coherent
   heaps. *)
Inductive heap_relation : Relation heap :=
| heap_relation_step : forall H1 H2,
  coherent_heap H1 ->
  heap_step H1 None H2 ->
  heap_relation H1 H2.

(* The reflexive, transitive closure of the above. *)
Definition heap_star := Rstar _ heap_relation.

(* If two heaps are related by the heap relation, they are both coherent. *)
Lemma heap_relation_implies_coherent : forall H1 H2,
  heap_relation H1 H2 ->
  coherent_heap H1 /\ coherent_heap H2.
Proof.
  intros H1 H2 relation.
  inversion relation; subst.
  split. auto. eapply empty_step_preserves_coherent. eauto. auto.
Qed.

(* The reflexive transitive closure of the above *)
Lemma heap_relation_star_preserves_coherent : forall H1 H2,
  coherent_heap H1 ->
  heap_star H1 H2 ->
  coherent_heap H2.
Proof.
  intros H1 H2 coherent star.
  induction star. auto.
  apply IHstar. eapply proj2. apply heap_relation_implies_coherent. eauto.
Qed.

(* The size of the buffer decreases with every step in the heap relation. Hence,
   the relation eventually terminates. *)
Lemma termination : forall S1 B1 Lambda1 S2 B2 Lambda2,
  heap_relation (S1, B1, Lambda1) (S2, B2, Lambda2) ->
  Buffer.size B1 = Buffer.size B2 + 1.
Proof.
  intros S1 B1 Lambda1 S2 B2 Lambda2 reln.
  inversion reln as [ H1 H2 coherent steps ].

  (* take an empty step *)
  subst H1 H2.
  inversion steps as [ | | | ? ? ? t r v ? B2_eq | | | | ]; subst.
  apply Buffer.dequeue_size in B2_eq.
  rewrite B2_eq. auto.
Qed.

(* This basically says that the heap relation terminates. *)
Lemma heap_relation_noetherian :
  Noetherian heap heap_relation.
Proof.
  intros H.
  destruct H as [ [ S B ] Lambda ].
  remember (Buffer.size B) as n.
  symmetry in Heqn.
  generalize dependent Lambda.
  generalize dependent B.
  generalize dependent S.
  induction n as [ | n ]; intros S B size_B Lambda;
    apply definition_of_noetherian;
      intros [ [ S' B' ] Lambda' ] reln;
          pose proof termination reln as size_B'.

  (* size B = 0 *)
  rewrite size_B in size_B'.
  rewrite plus_1 in size_B'.
  inversion size_B'.

  (* size B = n *)
  apply IHn. omega.
Qed.

(* Local confluence - if H1 steps to H2, and H1 also steps to H3, then there
   exists H4 such that H2 and H3 both step to H4 (possibly via many steps). *)
Lemma heap_relation_locally_confluent :
  Locally_confluent heap heap_relation.
Proof.
  intros H H1 H2 reln1 reln2.
  unfold coherent.
  inversion reln1 as [ ? ? coherent_H steps1 ].
  inversion reln2 as [ ? ? ? steps2 ]. subst.

  assert (H1 = H2 \/ exists H3, heap_step H1 None H3 /\ heap_step H2 None H3)
    as [ H1_eq_H2 | [ H3 [ steps13 steps23 ] ] ].

  (* Proving that an intermediate heap exists. *)
  destruct H as [ [ S B ] Lambda ].
  destruct H1 as [ [ S1 B1 ] Lambda1 ].
  destruct H2 as [ [ S2 B2 ] Lambda2 ].
  inversion steps1 as [ | | | ? ? ? t1 r1 v1 ? B1_eq | | | | ]; subst.
  inversion steps2 as [ | | | ? ? ? t2 r2 v2 ? B2_eq | | | | ]; subst.

  (* The dequeue_conflence lemma is expressed in terms of VarPair.eq *)
  destruct ((t1, r1) ==== (t2, r2)) as [ [ ? ? ] | Neq ].

  (* Case: t1 = t2 and r1 = r2 *)
  simpl in *. subst. rewrite B2_eq in B1_eq. inversion B1_eq; subst*.

  (* Case t1 = t2 or r1 = r2 *)
  pose proof Buffer.dequeue_confluence B1_eq B2_eq Neq
    as [ B' [ B'_eq1 B'_eq2 ] ].
  destruct (t1 == t2) as [ | Neq_t ].

  (* Case: t1 = t2 *)
  right. exists (Store.update (Store.update S r1 v1) r2 v2, B', Lambda2).
  split. apply* step_commit.
  rewrite Store.eq_update_update. apply* step_commit.
  intro false. apply Neq. unfold VarPair.eq. simpl. auto.

  (* Case: t1 = t2 *)
  destruct (r1 == r2).

  (* Case: t1 <> t2 and r1 = r2. Impossible by coherency. *)
  apply Buffer.dequeue_not_is_empty_queue in B1_eq.
  apply Buffer.dequeue_not_is_empty_queue in B2_eq.
  unfold coherent_heap, empty_except in coherent_H. simpl in coherent_H.
  pose proof coherent_H _ _ B1_eq _ Neq_t.
  subst. tauto.

  (* Case: t1 <> t2 and r1 <> r2. *)
  right. exists (Store.update (Store.update S r1 v1) r2 v2, B', Lambda2).
  split. apply* step_commit.
  rewrite* Store.eq_update_update. apply* step_commit.

  (* Intermediate heap exists. *)
  subst H2. exists H1. split; apply Rstar_0.
  exists H3. eauto 8 using Rstar_n, heap_relation_step,
  empty_step_preserves_coherent, Rstar_0.
Qed.

(* Global confluence: like local confluence, except now H2 and H3 may be many
   steps away from H1. *)
Lemma heap_relation_confluent :
  Confluent heap heap_relation.
Proof.
  apply Newman.
  apply heap_relation_noetherian.
  apply heap_relation_locally_confluent.
Qed.

(* An empty buffer cannot take any more steps. *)
Lemma empty_buffer_heap_relation_star : forall H1 H2,
  Buffer.is_empty (snd (fst H1)) ->
  heap_star H1 H2 ->
  H1 = H2.
Proof.
  intros H1 H2 Emp Star.
  rewrite Buffer.empty_size in Emp.
  induction Star as [ | H1 H2 H3 Reln ]; auto.
  destruct H1 as [ [ ? ? ] ? ].
  destruct H2 as [ [ ? ? ] ? ].
  apply termination in Reln. simpl in *.
  rewrite Emp in Reln. rewrite plus_comm in Reln.
  inversion Reln.
Qed.

(* The heap relation preserves the lock map. *)
Lemma heap_relation_implies_lock_maps_equal : forall H1 H2,
  heap_relation H1 H2 ->
  snd H1 = snd H2.
Proof.
  intros H1 H2 reln.
  inversion reln as [ ? ? ? steps ].
  inversion steps. simpl. auto.
Qed.

(* The reflexive, transitive closure of the above. *)
Lemma heap_relation_star_implies_lock_maps_equal : forall H1 H2,
  heap_star H1 H2 ->
  snd H1 = snd H2.
Proof.
  intros H1 H2 star.
  induction star as [ | ? ? ? reln ]; auto.
  erewrite heap_relation_implies_lock_maps_equal; eauto.
Qed.

(* The heap relation isn't affected by the lock map. *)
Lemma heap_relation_independent_of_lock_map : forall S B Lambda S' B' Lambda',
  heap_relation (S, B, Lambda) (S', B', Lambda) ->
  heap_relation (S, B, Lambda') (S', B', Lambda').
Proof.
  intros S B Lambda S' B' Lambda' reln.
  inversion reln as [ ? ? ? steps ].
  inversion steps. subst.
  apply heap_relation_step.
  apply coherent_independent_of_lock_map with (Lambda := Lambda). auto.
  apply step_commit with (t := t). auto.
Qed.

(* The reflexivie transitive closure of the above. *)
Lemma heap_relation_star_independent_of_lock_map :
  forall S B Lambda S' B' Lambda',
    heap_star (S, B, Lambda) (S', B', Lambda) ->
    heap_star (S, B, Lambda') (S', B', Lambda').
Proof.
  intros S B Lambda S' B' Lambda' star.
  unfold heap_star in *.
  remember (S, B, Lambda) as H.
  remember (S', B', Lambda) as H'.
  generalize dependent Lambda.
  generalize dependent B.
  generalize dependent S.
  induction star as [ | ? [ [ S'' B'' ] Lambda'' ] ? reln ];
    intros S B Lambda H_eq H'_eq; subst.
  inversion H'_eq. apply Rstar_0.
  pose proof heap_relation_implies_lock_maps_equal reln as eq.
  simpl in eq. subst.
  eapply Rstar_n.
  eapply heap_relation_independent_of_lock_map. eauto. eauto.
Qed.

(* The heap relation preserves heap validity. *)
Lemma heap_relation_preserves_well_typed : forall Sigma H1 H2,
  well_typed Sigma H1 ->
  heap_relation H1 H2 ->
  well_typed Sigma H2.
Proof.
  intros Sigma H1 H2 Typ reln.
  inversion reln as [ ? ? ? steps ]; subst.
  apply* preservation_empty.
Qed.

(* The reflexive transitive closure of the heap relation preserves heap
   validity. *)
Lemma heap_relation_star_preserves_well_typed : forall Sigma H1 H2,
  well_typed Sigma H1 ->
  heap_star H1 H2 ->
  well_typed Sigma H2.
Proof.
  intros Sigma H1 H2 Typ reln.
  induction reln. auto.
  eauto using heap_relation_preserves_well_typed.
Qed.

(* The heap relation preserves freshness for valid heaps. *)
Lemma heap_relation_preserves_fresh : forall Sigma H1 H2 r,
  well_typed Sigma H1 ->
  Store.fresh (fst (fst H1)) r ->
  heap_relation H1 H2 ->
  Store.fresh (fst (fst H2)) r.
Proof.
  intros Sigma H1 H2 r1 Typ Fresh Reln.
  inversion Reln as [ ? ? ? Step ].
  inversion Step as [ | | | ? ? ? t2 r2 ? ? Deq | | | | ]; subst.
  apply Buffer.dequeue_not_is_empty_queue in Deq.
  apply* Store.fresh_after_update. intro; subst.
  rewrite* <- well_typed_fresh_iff in Fresh.
  destruct Typ as [ Dom Typ ].
  specializes Dom (>>> Fresh). destruct Dom as [ Fresh2 Empty2 ].
  apply Deq. apply Empty2.
Qed.

(* Relfexive/transitive closure of the above. *)
Lemma heap_relation_star_preserves_fresh : forall Sigma H1 H2 r,
  well_typed Sigma H1 ->
  Store.fresh (fst (fst H1)) r ->
  heap_star H1 H2 ->
  Store.fresh (fst (fst H2)) r.
Proof.
  intros Sigma H1 H2 r Typ Fresh Star.
  induction* Star.
  eauto using heap_relation_preserves_fresh, heap_relation_preserves_well_typed.
Qed.

(* A "diamond lemma" for alloc steps.  Basically we push an alloc step after
   any empty step. *)
Lemma alloc_heap_relation_diamond : forall Sigma H1 H1' H2 t r v,
  well_typed Sigma H1 ->
  heap_relation H1 H2 ->
  heap_step H1 (Some (t, alloc r v)) H1' ->
  exists H2',
    heap_relation H1' H2' /\
    heap_step H2 (Some (t, alloc r v)) H2'.
Proof.
  intros Sigma H1 H1' H2 t r v typ1 reln steps1.
  inversion reln as [ ? ? coherent_H1 empty_step ]. subst.
  inversion empty_step as [ | | | S1 B1 Lambda1 t' r' v' B2 B2_eq | | | | ].
  subst.
  inversion steps1 as [ ? ? ? ? ? ? fresh1 | | | | | | | ]. subst.
  pose proof heap_relation_preserves_well_typed _ typ1 reln as typ2.
  assert (r <> r') as r_neq_r'. intro; subst r'.
  apply Buffer.dequeue_not_is_empty_queue in B2_eq.
  pose proof well_typed_fresh_buffer_empty typ1 fresh1 as empty1.
  apply B2_eq. apply empty1.

  (* H2' exists *)
  exists (Store.update (Store.update S1 r v) r' v', B2, Lambda1).

  (* H1' is coherent. *)
  split. apply heap_relation_step. unfold coherent_heap in *. auto.

  (* H1' steps to H2' via an empty step. *)
  eapply step_commit; eauto.

  (* H2 steps to H2' via an alloc. *)
  rewrite Store.eq_update_update; auto.
  apply step_alloc. apply Store.fresh_after_update; auto.
Qed.

(* The reflexive, transitive closure of the above. *)
Lemma alloc_heap_relation_star_diamond : forall Sigma H1 H1' H2 t r v,
  well_typed Sigma H1 ->
  heap_star H1 H2 ->
  heap_step H1 (Some (t, alloc r v)) H1' ->
  exists H2',
    heap_star H1' H2' /\
    heap_step H2 (Some (t, alloc r v)) H2'.
Proof.
  intros Sigma H1 H1' H2 t r v typ1 reln steps1.
  generalize dependent H1'.
  induction reln as [ H1 | H1 H3 H2 reln1 reln2 ]; intros H1' steps1.

  (* same heap *)
  exists H1'. split. apply Rstar_0. auto.

  (* inductive case - basically we're "pushing" the alloc to the end of the
     computation *)
  pose proof alloc_heap_relation_diamond _ typ1 reln1 steps1
    as [ H3' [ reln3 steps3 ] ].
  pose proof heap_relation_preserves_well_typed _ typ1 reln1 as valid3.
  pose proof IHreln2 valid3 _ steps3 as [ H2' [ reln4 steps2 ] ].
  exists H2'. split. eapply Rstar_n; eauto. auto.
Qed.

(* The heap relation preserves "empty except." *)
Lemma heap_relation_preserves_empty_except : forall H1 H2 t r,
  empty_except (snd (fst H1)) t r ->
  heap_relation H1 H2 ->
  empty_except (snd (fst H2)) t r.
Proof.
  intros H1 H2 t r empty_not_t reln t' t_neq_t'.
  inversion reln as [ ? ? ? steps ]; subst.
  inversion steps as [ | | | S1 B1 Lambda1 t1 r1 v1 B2 B2_eq | | | | ]; subst.
  simpl in *.
  eapply Buffer.is_empty_queue_dequeue; eauto.
Qed.

(* A "diamond lemma" for write steps - under certain circumstances, we can push
   a write step past any later empty steps. *)
Lemma write_heap_relation_diamond : forall H1 H1' H2 t r v,
  heap_relation H1 H2 ->
  empty_except (snd (fst H1)) t r ->
  heap_step H1 (Some (t, write r v)) H1' ->
  exists H2',
    heap_relation H1' H2' /\
    heap_step H2 (Some (t, write r v)) H2'.
Proof.
  intros H1 H1' H2 t r v reln except steps.
  inversion reln as [ ? ? coherent steps_empty ]; subst.
  pose proof write_preserves_coherent coherent except steps as coherent'.
  inversion steps_empty as [ | | | S1 B1 Lambda1 t1 r1 v1 B2 B2_eq | | | | ];
    subst.
  inversion steps; subst.
  pose proof Buffer.dequeue_enqueue t r v B2_eq as B3_eq.
  esplit. split.
  apply heap_relation_step. auto.
  eapply step_commit. eauto.
  apply step_write.
Qed.

(* The reflexive, transitive closure of the above. *)
Lemma write_heap_relation_star_diamond : forall H1 H1' H2 t r v,
  empty_except (snd (fst H1)) t r ->
  heap_star H1 H2 ->
  heap_step H1 (Some (t, write r v)) H1' ->
  exists H2',
    heap_star H1' H2' /\
    heap_step H2 (Some (t, write r v)) H2'.
Proof.
  intros H1 H1' H2 t r v empty_not_t reln steps1.
  generalize dependent H1'.
  induction reln as [ H1 | H1 H3 H2 reln1 reln2 ]; intros H1' steps1.

  (* same heap *)
  exists H1'. split. apply Rstar_0. auto.

  (* inductive case - basically we're "pushing" the alloc to the end of the
     computation *)
  pose proof write_heap_relation_diamond reln1 empty_not_t steps1
    as [ H3' [ reln3 steps3 ] ].
  pose proof heap_relation_preserves_empty_except empty_not_t reln1
    as empty_not_t'.
  pose proof IHreln2 empty_not_t' _ steps3 as [ H2' [ reln4 steps2 ] ].
  exists H2'. split. eapply Rstar_n; eauto. auto.
Qed.

(* The heap relation doesn't add any values to empty buffers *)
Lemma heap_relation_preserves_is_empty_queue : forall H1 H2 t r,
  Buffer.is_empty_queue (snd (fst H1)) t r ->
  heap_relation H1 H2 ->
  Buffer.is_empty_queue (snd (fst H2)) t r.
Proof.
  intros H1 H2 t r empty1 reln.
  inversion reln as [ ? ? coher steps ]; subst.
  inversion steps as [ | | | S1 B2 Lambda1 t1 r1 v1 B' B'_eq | | | | ]; subst.
  eapply Buffer.is_empty_queue_dequeue; eauto.
Qed.

(* If the buffer has no buffered values for r, then r's value in the store is
   preserved. *)
Lemma heap_relation_preserves_get : forall H1 H2 r v,
  Store.get (fst (fst H1)) r v ->
  Buffer.is_empty_loc (snd (fst H1)) r ->
  heap_relation H1 H2 ->
  Store.get (fst (fst H2)) r v.
Proof.
  intros H1 H2 r v get1 empty1 reln.
  inversion reln as [ ? ? coher steps ]; subst.
  inversion steps as [ | | | S1 B2 Lambda1 t1 r1 v1 B' B'_eq | | | | ]; subst.
  simpl in *.
  apply Store.get_after_update. auto.
  intro. subst r1.
  apply Buffer.dequeue_not_is_empty_queue in B'_eq.
  apply B'_eq. apply empty1.
Qed.

(* Closure of the above. *)
Lemma heap_relation_star_preserves_get : forall H1 H2 r v,
  Store.get (fst (fst H1)) r v ->
  Buffer.is_empty_loc (snd (fst H1)) r ->
  heap_star H1 H2 ->
  Store.get (fst (fst H2)) r v.
Proof.
  intros H1 H2 r v get1 empty1 star.
  induction star. auto.
  apply IHstar.
  eapply heap_relation_preserves_get; eauto.
  unfold Buffer.is_empty_loc in *.
  intros t1. pose proof empty1 t1.
  eapply heap_relation_preserves_is_empty_queue; eauto.
Qed.

(* The heap relation does not any new locations to the store (although of
   course it may change the values of existing locations. *)
Lemma heap_relation_preserves_get_exists : forall H1 H2 r v,
  heap_relation H1 H2 ->
  Store.get (fst (fst H1)) r v ->
  exists v', Store.get (fst (fst H2)) r v'.
Proof.
  intros H1 H2 r1 v1 reln gets.
  inversion reln as [ ? ? ? steps ]. subst.
  inversion steps as [ | | | S1 B1 Lambda1 t2 r2 v2 B2 B2_eq | | | | ]. subst.
  simpl in *. destruct (r1 == r2). subst.
  exists v2. apply Store.get_update_same.
  exists v1. apply* Store.get_after_update.
Qed.

(* Closure of the above *)
Lemma heap_relation_star_preserves_get_exists : forall H1 H2 r v,
  heap_star H1 H2 ->
  Store.get (fst (fst H1)) r v ->
  exists v', Store.get (fst (fst H2)) r v'.
Proof.
  intros H1 H2 r v star get1.
  generalize dependent v.
  induction star; intros v get1. eauto.
  forwards get2 : heap_relation_preserves_get_exists (>>> __); eauto.
  destruct* get2 as [ ? ? ].
Qed.

Inductive top_value : heap -> loc -> L.val -> Prop :=
| top_value_top : forall S B L t r v,
  empty_except B t r ->
  Buffer.top B t r = Some v ->
  top_value (S, B, L) r v
| top_value_get : forall S B L r v,
  Buffer.is_empty_loc B r ->
  Store.get S r v ->
  top_value (S, B, L) r v.

Lemma heap_relation_preserves_top_value : forall H1 H2 r v,
  heap_relation H1 H2 ->
  top_value H1 r v ->
  top_value H2 r v.
Proof with eauto.
  intros H1 H2 r v reln top1.
  inversion reln as [ ? ? coher steps ]; subst.
  inversion steps as [ | | | S1 B1 Lambda1 t' r' v' B2 B2_eq | | | | ]; subst.
  inversion top1 as [ ? ? ? t ? ? except top_eq | ? ? ? ? ? empty_B get_S ];
    subst.
  forwards except' : heap_relation_preserves_empty_except (>>>Hyps __ reln).
  simpl... simpl in except'.
  destruct (VarPair.eq_dec (t, r) (t', r')) as [ [ ? ? ] | ].
  simpl in *. subst.

  forwards [ [ ? ? ] | ? ] : Buffer.top_dequeue (>>> top_eq B2_eq).
  subst.
  apply top_value_get.
  intros t''.
  destruct (t' == t'')... subst...
  apply Store.get_update_same.

  eapply top_value_top...

  eapply top_value_top...
  eapply Buffer.top_dequeue_diff...

  destruct (r == r'). subst.

  apply Buffer.dequeue_not_is_empty_queue in B2_eq.
  contradiction B2_eq...

  apply top_value_get.
  intros t.
  eapply Buffer.is_empty_queue_dequeue...
  apply Store.get_after_update...
Qed.

Lemma heap_relation_star_preserves_top_value : forall H1 H2 r v,
  heap_star H1 H2 ->
  top_value H1 r v ->
  top_value H2 r v.
Proof.
  intros H1 H2 r v star top1.
  generalize dependent v.
  induction star; intros v top1.
  auto.
  forwards top2 : heap_relation_preserves_top_value (>>> __); eauto.
Qed.
