Require Import Metatheory.

Require Import models.
Require Import var_map.

Module Lambda_Calculus <: LANGUAGE_MODEL.
  (** Grammar of terms. *)
  Inductive trm : Set :=
  | trm_unit  : trm
  | trm_bvar  : nat -> trm
  | trm_fvar  : var -> trm
  | trm_abs   : trm -> trm
  | trm_app   : trm -> trm -> trm
  | trm_loc   : loc -> trm
  | trm_alloc : trm -> trm
  | trm_get   : trm -> trm
  | trm_set   : trm -> trm -> trm
  | trm_lock  : lock -> trm
  | trm_acq   : trm -> trm
  | trm_rel   : trm -> trm
  | trm_spawn : trm -> trm.

  (* Terms are stored in the heap. *)
  Definition val := trm.

  (** Definition of values (deliberately syntactic - I don't want to reason
     about local closure etc.) *)
  Inductive value : trm -> Prop :=
  | value_unit :
    value trm_unit
  | value_abs : forall t1,
    value (trm_abs t1)
  | value_loc : forall r,
    value (trm_loc r)
  | value_lock : forall l,
    value (trm_lock l).

  (** Opening up abstractions *)
  Fixpoint open_rec (k : nat) (u : trm) (t : trm) {struct t} : trm :=
    match t with
      | trm_unit      => trm_unit
      | trm_bvar i    => if k === i then u else (trm_bvar i)
      | trm_fvar x    => trm_fvar x
      | trm_abs t1    => trm_abs (open_rec (S k) u t1)
      | trm_app t1 t2 => trm_app (open_rec k u t1) (open_rec k u t2)
      | trm_loc r     => trm_loc r
      | trm_alloc t1  => trm_alloc (open_rec k u t1)
      | trm_get t1    => trm_get (open_rec k u t1)
      | trm_set t1 t2 => trm_set (open_rec k u t1) (open_rec k u t2)
      | trm_lock l    => trm_lock l
      | trm_acq t1    => trm_acq (open_rec k u t1)
      | trm_rel t1    => trm_rel (open_rec k u t1)
      | trm_spawn t1  => trm_spawn (open_rec k u t1)
    end.

  Definition open t u := open_rec 0 u t.

  Notation "{ k ~> u } t" := (open_rec k u t) (at level 67).
  Notation "t ^^ u" := (open t u) (at level 67).
  Notation "t ^ x" := (open t (trm_fvar x)).

  (** Reduction relation - one step in call-by-value *)
  Inductive red : trm -> action trm -> trm -> option trm -> Prop :=
  | red_beta : forall t1 t2,
    value t2 ->
    red (trm_app (trm_abs t1) t2) (pure _) (t1 ^^ t2) None
  | red_app_1 : forall t1 a t1' t2 t_opt,
    red t1 a t1' t_opt ->
    red (trm_app t1 t2) a (trm_app t1' t2) t_opt
  | red_app_2 : forall t1 t2 a t2' t_opt,
    value t1 ->
    red t2 a t2' t_opt ->
    red (trm_app t1 t2) a (trm_app t1 t2') t_opt
  | red_alloc : forall t1 r,
    value t1 ->
    red (trm_alloc t1) (alloc r t1) (trm_loc r) None
  | red_alloc_1 : forall t1 a t1' t_opt,
    red t1 a t1' t_opt ->
    red (trm_alloc t1) a (trm_alloc t1') t_opt
  | red_get : forall r t1,
    red (trm_get (trm_loc r)) (read r t1) t1 None
  | red_get_1 : forall t1 a t1' t_opt,
    red t1 a t1' t_opt ->
    red (trm_get t1) a (trm_get t1') t_opt
  | red_set : forall r t1,
    value t1 ->
    red (trm_set (trm_loc r) t1) (write r t1) trm_unit None
  | red_set_1 : forall t1 a t1' t2 t_opt,
    red t1 a t1' t_opt ->
    red (trm_set t1 t2) a (trm_set t1' t2) t_opt
  | red_set_2 : forall t1 t2 a t2' t_opt,
    value t1 ->
    red t2 a t2' t_opt ->
    red (trm_set t1 t2) a (trm_set t1 t2') t_opt
  | red_acq : forall l,
    red (trm_acq (trm_lock l)) (acquire _ l) trm_unit None
  | red_acq_1 : forall t1 a t1' t_opt,
    red t1 a t1' t_opt ->
    red (trm_acq t1) a (trm_acq t1') t_opt
  | red_rel : forall l,
    red (trm_rel (trm_lock l)) (release _ l) trm_unit None
  | red_rel_1 : forall t1 a t1' t_opt,
    red t1 a t1' t_opt ->
    red (trm_rel t1) a (trm_rel t1') t_opt
  | red_spawn : forall t1 id,
    red (trm_spawn t1) (spawn _ id) trm_unit (Some t1).

  Notation "t -- a --> t' ;; t_opt" := (red t a t' t_opt) (at level 68).

  (* Program states *)
  Definition program_state := Map.t trm.

  (* Program steps *)
  Inductive step : program_state -> effect val -> program_state -> Prop :=
  | step_thread : forall P id t a t',
    Map.MapsTo id t P ->
    t -- a --> t' ;; None ->
    step P (Some (id, a)) (Map.add id t' P)
  | step_spawn : forall P id1 t1 t1' id2 t2,
    Map.MapsTo id1 t1 P ->
    ~ Map.In id2 P ->
    t1 -- (spawn _ id2) --> t1' ;; (Some t2) ->
    step P (Some (id1, spawn _ id2)) (Map.add id2 t2 (Map.add id1 t1' P)).

  Definition program_step := step.

  (** Reflexive transitive closure of program step. *)
  Inductive program_step_many : program_state -> program_state -> trace val
    -> Prop :=
  | program_step_many_refl : forall P,
    program_step_many P P nil
  | program_step_many_trans : forall P1 P2 P3 tr e,
    program_step_many P1 P2 tr ->
    program_step P2 e P3 ->
    program_step_many P1 P3 (e :: tr).

  Lemma no_empty_steps : forall P P',
    ~ (program_step P None P').
  Proof.
    introv Step. inversion Step.
  Qed.
End Lambda_Calculus.
