Require Import Metatheory.

Require Import lambda_calculus.
Require Import models.
Require Import types.
Require Import var_map.

(** This is a very simple type system that guarantees the program state does
   not contain any unallocated locations. *)
Module Locations_Ok <: TYPE_MODEL.
  Module L := Lambda_Calculus.
  Import L.

  (** There's only one type. *)
  Definition type := unit.

  (** A heap typing is a map from locations to types. *)
  Definition heap_typing := Map.t type.
  Definition empty_heap_typing := Map.empty type.

  (** Sigma' extends Sigma if it contains all of Sigma's mapping. *)
  Definition extends (Sigma Sigma' : heap_typing) :=
    forall r T, Map.MapsTo r T Sigma ->
      Map.MapsTo r T Sigma'.
  Hint Unfold extends.

  (** A well-typed term has no unknown locations. *)
  Inductive locations_ok : heap_typing -> trm -> Prop :=
  | locations_ok_unit  : forall Sigma,
    locations_ok Sigma trm_unit
  | locations_ok_bvar  : forall Sigma n,
    locations_ok Sigma (trm_bvar n)
  | locations_ok_fvar  : forall Sigma x,
    locations_ok Sigma (trm_fvar x)
  | locations_ok_abs   : forall Sigma t1,
    locations_ok Sigma t1 ->
    locations_ok Sigma (trm_abs t1)
  | locations_ok_app   : forall Sigma t1 t2,
    locations_ok Sigma t1 ->
    locations_ok Sigma t2 ->
    locations_ok Sigma (trm_app t1 t2)
  | locations_ok_loc   : forall Sigma r,
    Map.MapsTo r tt Sigma ->
    locations_ok Sigma (trm_loc r)
  | locations_ok_alloc : forall Sigma t1,
    locations_ok Sigma t1 ->
    locations_ok Sigma (trm_alloc t1)
  | locations_ok_get   : forall Sigma t1,
    locations_ok Sigma t1 ->
    locations_ok Sigma (trm_get t1)
  | locations_ok_set   : forall Sigma t1 t2,
    locations_ok Sigma t1 ->
    locations_ok Sigma t2 ->
    locations_ok Sigma (trm_set t1 t2)
  | locations_ok_lock  : forall Sigma l,
    locations_ok Sigma (trm_lock l)
  | locations_ok_acq   : forall Sigma t1,
    locations_ok Sigma t1 ->
    locations_ok Sigma (trm_acq t1)
  | locations_ok_rel   : forall Sigma t1,
    locations_ok Sigma t1 ->
    locations_ok Sigma (trm_rel t1)
  | locations_ok_spawn : forall Sigma t1,
    locations_ok Sigma t1 ->
    locations_ok Sigma (trm_spawn t1).

  Hint Constructors locations_ok.

  (** Typing for terms. *)
  Inductive typing_trm : heap_typing -> trm -> type -> Prop :=
  | typing_trm_1 : forall Sigma t1,
    locations_ok Sigma t1 ->
    typing_trm Sigma t1 tt.

  Hint Constructors typing_trm.

  (** Term typing <-> locations_ok. *)
  Lemma typing_trm_inversion : forall Sigma t1,
    typing_trm Sigma t1 tt <-> locations_ok Sigma t1.
  Proof.
    split.
    introv Typ. inversions* Typ.
    intros. apply* typing_trm_1.
  Qed.

  (** Typing judgments for values and program states. *)
  Definition typing_val := typing_trm.
  Definition typing (Sigma : heap_typing) (P : program_state) : Prop :=
    forall id t1,
      Map.MapsTo id t1 P ->
      locations_ok Sigma t1.
  Hint Unfold typing_val typing.

  Lemma locations_ok_stability : forall Sigma t Sigma',
    locations_ok Sigma t ->
    extends Sigma Sigma' ->
    locations_ok Sigma' t.
  Proof.
    introv WF Ext.
    induction* WF.
  Qed.

  Hint Resolve locations_ok_stability.

  (** Value typing holds for an extending heap typing. *)
  Lemma typing_val_stability : forall Sigma v T Sigma',
    typing_val Sigma v T ->
    extends Sigma Sigma' ->
    typing_val Sigma' v T.
  Proof.
    introv Typ Ext.
    induction T.
    apply typing_trm_inversion in Typ.
    apply* typing_trm_1.
  Qed.
  (** The heap must ensure that any values read are well-typed and that any
     locations allocated are fresh. *)
  Inductive typing_incoming : heap_typing -> action L.val -> Prop :=
  | typing_incoming_alloc : forall Sigma r v,
    ~ Map.In r Sigma ->
    typing_incoming Sigma (alloc r v)
  | typing_incoming_read : forall Sigma r v T,
    Map.MapsTo r T Sigma ->
    typing_val Sigma v T ->
    typing_incoming Sigma (read r v)

  | typing_incoming_write : forall Sigma r v,
    typing_incoming Sigma (write r v)
  | typing_incoming_acquire : forall Sigma l,
    typing_incoming Sigma (acquire _ l)
  | typing_incoming_release : forall Sigma l,
    typing_incoming Sigma (release _ l)
  | typing_incoming_spawn : forall Sigma id,
    typing_incoming Sigma (spawn _ id).

  (** The program must ensure that any values written or allocated are
     well-typed. *)
  Inductive typing_outgoing :
    heap_typing -> action L.val -> heap_typing -> Prop :=
  | typing_outgoing_alloc : forall Sigma r v T,
    typing_val Sigma v T ->
    typing_outgoing Sigma (alloc r v) (Map.add r T Sigma)
  | typing_outgoing_write : forall Sigma r v T,
    Map.MapsTo r T Sigma ->
    typing_val Sigma v T ->
    typing_outgoing Sigma (write r v) Sigma

  | typing_outgoing_read : forall Sigma r v,
    typing_outgoing Sigma (read r v) Sigma
  | typing_outgoing_acquire : forall Sigma l,
    typing_outgoing Sigma (acquire _ l) Sigma
  | typing_outgoing_release : forall Sigma l,
    typing_outgoing Sigma (release _ l) Sigma
  | typing_outgoing_spawn : forall Sigma id,
    typing_outgoing Sigma (spawn _ id) Sigma.

  Hint Constructors typing_incoming typing_outgoing.
End Locations_Ok.

(** This module guarantees that the Locations_Ok type system is preserved by
   the language semantics (assuming the memory model is type-preserving. *)
Module Locations_Ok_Preservation <: TYPE_MODEL_PRESERVATION.
  Module L := Lambda_Calculus.
  Module Typ := Locations_Ok.
  Import Lambda_Calculus.
  Import Locations_Ok.

  Module TypeFacts := Type_Model_Facts Typ.
  Module Facts := FMapFacts.Facts Map.
  Hint Resolve TypeFacts.typing_extends.
  Hint Unfold Var_as_OT.eq.

  (* Tactic for the inductive cases in the preservation_red proof. *)
  Ltac pres_red_ind H := specializes* H (>>> __ __ __);
    destruct* H as [ ? [ ? ? ] ];
      esplit; auto*.

  (* If you have a variable of type T, it must be tt. *)
  Ltac type_to_tt :=
    match goal with
      | T : type |- _ => induction T
    end.

  (** Type preservation for non-spawn reductions. *)
  Lemma preservation_red : forall Sigma t1 a t1',
    locations_ok Sigma t1 ->
    red t1 a t1' None ->
    typing_incoming Sigma a ->
    exists Sigma',
      typing_outgoing Sigma a Sigma' /\
      locations_ok Sigma' t1'.
  Proof.
    introv WF. inductions WF gen t1' end;
      introv Red Inc; inversions Red.
    inversion Inc.
    pres_red_ind IHWF1.
    pres_red_ind IHWF2.
    exists (Map.add r tt Sigma). split*.
      apply locations_ok_loc. apply* Map.add_1.
    pres_red_ind IHWF.
    exists Sigma. split*.
      inversions Inc. type_to_tt.
      apply* typing_trm_inversion.
    pres_red_ind IHWF.
    inversions WF1. exists* Sigma.
    pres_red_ind IHWF1.
    pres_red_ind IHWF2.
    exists* Sigma.
    pres_red_ind IHWF.
    exists* Sigma.
    pres_red_ind IHWF.
  Qed.

  (* Tactic for the inductive cases in the preservation_spawn proof. *)
  Ltac pres_spawn_ind H := specializes* H (>>> __ __).

  (** Type preservation for spawn reductions. *)
  Lemma preservation_spawn : forall Sigma t1 id t1' t2,
    locations_ok Sigma t1 ->
    red t1 (spawn _ id) t1' (Some t2) ->
    locations_ok Sigma t1' /\ locations_ok Sigma t2.
  Proof.
    introv Typ. inductions Typ gen t1' end;
      introv Red; inversions Red.
    pres_spawn_ind IHTyp1.
    pres_spawn_ind IHTyp2.
    pres_spawn_ind IHTyp.
    pres_spawn_ind IHTyp.
    pres_spawn_ind IHTyp1.
    pres_spawn_ind IHTyp2.
    pres_spawn_ind IHTyp.
    pres_spawn_ind IHTyp.
    split*.
  Qed.

  Lemma preservation_read : forall Sigma t1 r v t1',
    locations_ok Sigma t1 ->
    red t1 (read r v) t1' None ->
      Map.In r Sigma.
  Proof.
    introv Typ. inductions Typ gen t1' end;
      introv Red; inversions Red; eauto.
    inversion Typ.
    esplit. eauto.
  Qed.

  Lemma preservation_write : forall Sigma t1 r v t1',
    locations_ok Sigma t1 ->
    red t1 (write r v) t1' None ->
      Map.In r Sigma.
  Proof.
    introv Typ. inductions Typ gen t1' end;
      introv Red; inversions Red; eauto.
    inversion Typ1.
    esplit. eauto.
  Qed.

  (** Typing is preserved by an extension of store typing. *)
  Lemma typing_stability : forall Sigma P Sigma',
    typing Sigma P ->
    extends Sigma Sigma' ->
    typing Sigma' P.
  Proof.
    intros Sigma P Sigma' Typ Ext id t Maps.
    unfold typing in Typ.
    specializes* Typ (>>>Hyps __).
  Qed.

  Hint Resolve typing_stability.

  (** Adding a new well-typed thread preserves typing for the program state. *)
  Lemma typing_add : forall Sigma P t id,
    typing Sigma P ->
    locations_ok Sigma t ->
    typing Sigma (Map.add id t P).
  Proof.
    intros Sigma P t id Typ1 Typ2 id' t' Maps.
    rewrite Facts.add_mapsto_iff in Maps.
    destruct Maps as [ [ ] | [ ? Maps ] ].
    subst. eauto.
    unfold typing in Typ1.
    apply* Typ1.
  Qed.

  Hint Resolve typing_add.

  (** Progress does not hold (due to locking). Preservation only holds in
     combination with a type-preserving heap semantics. This axiom guarantees
     that the program has type preservation for mutual steps. *)
  Lemma preservation : forall Sigma P id a P',
    typing Sigma P ->
    L.program_step P (Some (id, a)) P' ->
    typing_incoming Sigma a ->
    exists Sigma',
      typing_outgoing Sigma a Sigma' /\
      typing Sigma' P'.
  Proof.
    introv Typ Step Inc.
    unfold typing in Typ.
    inversions* Step.

    (* non-spawn step *)
    forwards* Typ1 : Typ.
    forwards Pres : preservation_red; eauto.
    destruct Pres as [ Sigma' [ ] ].
    forwards Ext : TypeFacts.typing_extends; eauto.

    (* spawn step *)
    forwards* Typ1 :Typ.
    forwards Pres : preservation_spawn; eauto.
    destruct* Pres as [ ? ? ].
  Qed.

  (** Well-formed is preserved by substitution. *)
  Lemma locations_ok_subst : forall Sigma t1 t2,
    locations_ok Sigma t1 ->
    locations_ok Sigma t2 ->
    locations_ok Sigma (t1 ^^ t2).
  Proof.
    introv Typ1 Typ2. unfold open.
    remember 0 as n. clear Heqn.
    inductions Typ1 gen n end; simpl; auto.
    intro n'. destruct* (n' === n).
  Qed.

  (** Type preservation for pure reductions. *)
  Lemma preservation_red_pure : forall Sigma t1 t1',
    locations_ok Sigma t1 ->
    red t1 (pure _) t1' None ->
    locations_ok Sigma t1'.
  Proof.
    introv Typ. inductions Typ gen t1' end;
      introv Red; inversions Red; auto*.
    inversions Typ1. apply* locations_ok_subst.
  Qed.

  (** Type preservation for pure steps. *)
  Lemma preservation_pure : forall Sigma P id P',
    typing Sigma P ->
    L.program_step P (Some (id, pure _)) P' ->
    typing Sigma P'.
  Proof.
    introv Typ Step.
    inversions* Step.
    unfold typing in Typ.
    forwards* Typ1 : Typ.
    forwards Pres : preservation_red_pure; eauto.
  Qed.
End Locations_Ok_Preservation.
