(** We say that two actions are "equal modulo values" if they are both reads,
   both writes, and so on, and they both act on the same location, but possibly
   have a different value. We extend this notion to erased traces (i.e., lists
   of TID/action pairs). *)

Require Import List.

Require Import actions.
Require Import indexes.

Set Implicit Arguments.

Section Traces.
  Variable val : Set.

  Inductive actions_equal_modulo_values : action val -> action val -> Prop :=
  | actions_equal_modulo_values_pure :
    actions_equal_modulo_values (pure _) (pure _)
  | actions_equal_modulo_values_alloc : forall r v1 v2,
    actions_equal_modulo_values (alloc r v1) (alloc r v2)
  | actions_equal_modulo_values_read : forall r v1 v2,
    actions_equal_modulo_values (read r v1) (read r v2)
  | actions_equal_modulo_values_write : forall r v1 v2,
    actions_equal_modulo_values (write r v1) (write r v2)
  | actions_equal_modulo_values_acquire : forall l,
    actions_equal_modulo_values (acquire _ l) (acquire _ l)
  | actions_equal_modulo_values_release : forall l,
    actions_equal_modulo_values (release _ l) (release _ l)
  | actions_equal_modulo_values_spawn : forall t,
    actions_equal_modulo_values (spawn _ t) (spawn _ t).

  Hint Constructors actions_equal_modulo_values.

  Lemma actions_equal_modulo_values_refl : forall a,
    actions_equal_modulo_values a a.
  Proof.
    intro a. induction a; auto.
  Qed.

  Lemma actions_equal_modulo_values_sym : forall a1 a2,
    actions_equal_modulo_values a1 a2 ->
    actions_equal_modulo_values a2 a1.
  Proof.
    intros a1 a2 mod.
    induction mod; auto.
  Qed.

  Lemma actions_equal_modulo_values_race : forall a1 a1' a2 a2',
    data_race a1 a2 ->
    actions_equal_modulo_values a1 a1' ->
    actions_equal_modulo_values a2 a2' ->
    data_race a1' a2'.
  Proof.
    intros a1 a1' a2 a2' race12 mod1 mod2.
    inversion race12; subst;
      inversion mod1; inversion mod2; subst.
    apply data_race_rw. apply data_race_wr. apply data_race_ww.
  Qed.

  Inductive equal_modulo_values : erased_trace val -> erased_trace val ->
    Prop :=
  | equal_modulo_values_nil : equal_modulo_values nil nil
  | equal_modulo_values_cons : forall tr1 tr2 t a1 a2,
    equal_modulo_values tr1 tr2 ->
    actions_equal_modulo_values a1 a2 ->
    equal_modulo_values ((t, a1) :: tr1) ((t, a2) :: tr2).

  Hint Constructors equal_modulo_values.

  Lemma equal_modulo_values_refl : forall tr,
    equal_modulo_values tr tr.
  Proof.
    intros tr.
    induction tr as [ | [ ? ? ] ];
      eauto using actions_equal_modulo_values_refl.
  Qed.

  Lemma equal_modulo_values_sym : forall tr1 tr2,
    equal_modulo_values tr1 tr2 ->
    equal_modulo_values tr2 tr1.
  Proof.
    intros tr1 tr2 mod.
    induction mod; eauto using actions_equal_modulo_values_sym.
  Qed.

  Lemma equal_modulo_values_index : forall tr1 tr2 j t a1,
    index tr1 j (t, a1) ->
    equal_modulo_values tr1 tr2 ->
    exists a2, index tr2 j (t, a2) /\ actions_equal_modulo_values a1 a2.
  Proof.
    intros tr1 tr2 j t a1 idx mod.
    remember (t, a1) as e1.
    generalize dependent tr2.
    induction idx as [ tr1 [ t3 a3 ] | ]; intros tr2 mod.
    inversion Heqe1; subst.
    inversion mod; subst. eauto.
    inversion mod as [ | ? tr2' ? ? ? mod' ]; subst.
    pose proof IHidx (refl_equal _) tr2' mod' as [ a2 [ idx2 actions_mod ] ].
    eauto.
  Qed.
End Traces.