(** This file implements the Modified-Shared-Invalid (MSI) cache coherency
   protocol as a memory model in our framework. *)

Require Import Metatheory.

Require Import maps.
Require Import models.

Module MSI (L : LANGUAGE_MODEL) <: MEMORY_MODEL L.

  (* store is a map from locations to values. *)
  Definition store := VarMap.t L.val.

  (* cache is a map from TID/location pairs to values tagged with cache
     states. *)
  Inductive cache_state := modified | shared.
  Definition cache := VarPairMap.t (cache_state * L.val).

  (* heap consists of a store and a cache *)
  Definition heap := store * cache.

  (* the empty heap *)
  Definition empty_heap : heap := (VarMap.empty _, VarPairMap.empty _).

  (* heap steps - must maintain cache coherency *)
  Inductive step : heap -> effect L.val -> heap -> Prop :=
  | step_alloc : forall S C t r v,
    ~ VarMap.In r S ->
    step (S, C) (Some (t, alloc r v)) (VarMap.add r v S, C)
  | step_read_shared : forall S C t r v,
    VarPairMap.MapsTo (t, r) (shared, v) C ->
    step (S, C) (Some (t, read r v)) (S, C)
  | step_read_modified : forall S C t r v,
    VarPairMap.MapsTo (t, r) (modified, v) C ->
    step (S, C) (Some (t, read r v)) (S, C)
  | step_write_modified : forall S C t r v v',
    VarPairMap.MapsTo (t, r) (modified, v) C ->
    step (S, C) (Some (t, write r v'))
    (S, VarPairMap.add (t, r) (modified, v') C)
  | step_invalid_to_shared : forall S C t r v,
    ~ VarPairMap.In (t, r) C ->
    (forall t' v', t <> t' ->
      ~ VarPairMap.MapsTo (t', r) (modified, v') C) ->
    VarMap.MapsTo r v S ->
    step (S, C) None (S, VarPairMap.add (t, r) (shared, v) C)
  | step_invalid_to_modified : forall S C t r v,
    (forall t', ~ VarPairMap.In (t', r) C) ->
    VarMap.MapsTo r v S ->
    step (S, C) None (S, VarPairMap.add (t, r) (modified, v) C)
  | step_shared_to_invalid : forall S C t r v,
    VarPairMap.MapsTo (t, r) (shared, v) C ->
    step (S, C) None (S, VarPairMap.remove (t, r) C)
  | step_modified_to_invalid : forall S C t r v,
    VarPairMap.MapsTo (t, r) (modified, v) C ->
    step (S, C) None
    (VarMap.add r v S, VarPairMap.remove (t, r) C)
  | step_spawn : forall H t1 t2,
    step H (Some (t1, spawn _ t2)) H.

  Definition heap_step := step.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).

  Lemma no_pure_steps : forall H H' t,
    ~ (heap_step H (Some (t, pure _)) H').
  Proof.
    intros H H' t Step. inversion Step.
  Qed.
End MSI.
