Require Import Arith.
Require Import Decidable.
Require Import List.

Require Import nat_facts.

Set Implicit Arguments.

Module Type FIFO_QUEUE.
  Parameter t : Type -> Type.
  Section Types.
    Variable elt : Type.
    Parameter empty : t elt.
    Parameter is_empty : t elt -> Prop.
    Parameter enqueue : t elt -> elt -> t elt.
    Parameter dequeue : t elt -> option (elt * t elt).
    Parameter size : t elt -> nat.
    Parameter top : t elt -> option elt.

    (* Specification *)
    Axiom empty_is_empty : is_empty empty.
    Axiom empty_dequeue : forall (q : t elt),
      is_empty q ->
      dequeue q = None.
    Axiom empty_size : forall q,
      is_empty q <-> size q = 0.
    Axiom dequeue_size : forall q e q',
      dequeue q = Some (e, q') ->
      size q = size q' + 1.
    Axiom is_empty_dec : forall q,
      decidable (is_empty q).
    Axiom enqueue_top : forall q e,
      top (enqueue q e) = Some e.
    Axiom dequeue_top : forall q e,
      top q = Some e ->
      exists e', exists q',
        dequeue q = Some (e', q') /\
        (top q' = Some e \/ (e = e' /\ is_empty q')).
    Axiom dequeue_enqueue_empty : forall q e,
      is_empty q ->
      exists q', dequeue (enqueue q e) = Some (e, q')
        /\ is_empty q'.
    Axiom not_empty_dequeue : forall q,
      ~ is_empty q ->
      exists e, exists q',
        dequeue q = Some (e, q').
    Axiom not_empty_enqueue : forall q v,
      ~ is_empty (enqueue q v).
    Axiom dequeue_enqueue : forall q q' v v',
      dequeue q = Some (v, q') ->
      dequeue (enqueue q v') = Some (v, enqueue q' v').
  End Types.
End FIFO_QUEUE.

Module ListQueue : FIFO_QUEUE.
  Definition t := list.
  Section Types.
    Variable elt : Type.

    Definition empty := nil (A := elt).

    Definition is_empty (q : t elt) := q = empty.

    Definition enqueue (q : t elt) (e : elt) := q ++ (e :: nil).

    Definition dequeue (q : t elt) :=
      match q with
        | nil => None
        | e::q' => Some (e, q')
      end.

    Definition size (q : t elt) := length q.

    Fixpoint top (q : t elt) : option elt :=
      match q with
        | nil => None
        | e :: nil => Some e
        | _ :: q' => top q'
      end.

    Fact empty_is_empty : is_empty empty.
    Proof.
      unfold is_empty. auto.
    Qed.

    Fact empty_dequeue : forall q,
      is_empty q ->
      dequeue q = None.
    Proof.
      intros q empty_q.
      unfold is_empty in *. subst.
      simpl. auto.
    Qed.

    Fact empty_size : forall q,
      is_empty q <-> size q = 0.
    Proof.
      intros q. unfold is_empty, empty. split.
      intros. subst. auto.
      intros size_q. destruct q. auto. simpl in *. inversion size_q.
    Qed.

    Fact dequeue_size : forall q e q',
      dequeue q = Some (e, q') ->
      size q = size q' + 1.
    Proof.
      intros q e q' q'_eq.
      destruct q.
      inversion q'_eq.
      simpl in q'_eq. inversion q'_eq. subst.
      simpl. symmetry. apply plus_1.
    Qed.

    Fact is_empty_dec : forall q,
      decidable (is_empty q).
    Proof.
      intro q. unfold decidable.
      destruct q.
      left. apply empty_is_empty.
      right. intro false.
      unfold is_empty in false. inversion false.
    Qed.

    Fact enqueue_top : forall q e,
      top (enqueue q e) = Some e.
    Proof.
      intros q e.
      unfold enqueue.
      induction q.
      simpl. auto.
      rewrite <- app_comm_cons.
      simpl. destruct (q ++ e :: nil); auto.
      simpl in IHq. inversion IHq.
    Qed.

    Fact dequeue_top : forall q e,
      top q = Some e ->
      exists e', exists q',
        dequeue q = Some (e', q') /\
        (top q' = Some e \/ (e = e' /\ is_empty q')).
    Proof.
      intros q e top_eq.
      induction q as [ | e' q' ].
      simpl in top_eq. inversion top_eq.
      exists e'. exists q'.
      split. simpl. auto.
      induction q' as [ | e'' q'' ].
      right. simpl in top_eq. inversion top_eq.
      split. auto. apply empty_is_empty.
      left. simpl in top_eq. simpl. auto.
    Qed.

    Lemma dequeue_enqueue_empty : forall q e,
      is_empty q ->
      exists q', dequeue (enqueue q e) = Some (e, q')
        /\ is_empty q'.
    Proof.
      intros q e empty_q.
      unfold is_empty, empty in *. subst q.
      unfold enqueue, dequeue. simpl. eauto.
    Qed.

    Lemma not_empty_dequeue : forall q,
      ~ is_empty q ->
      exists e, exists q',
        dequeue q = Some (e, q').
    Proof.
      intros q not_empty.
      destruct q as [ | e q ].
      unfold is_empty, empty in not_empty. tauto.
      exists e. exists q. simpl. auto.
    Qed.

    Lemma not_empty_enqueue : forall q v,
      ~ is_empty (enqueue q v).
    Proof.
      intros q v.
      unfold enqueue.
      rewrite empty_size.
      unfold size.
      rewrite app_length. simpl.
      rewrite plus_1. auto.
    Qed.

    Lemma dequeue_enqueue : forall q q' v v',
      dequeue q = Some (v, q') ->
      dequeue (enqueue q v') = Some (v, enqueue q' v').
    Proof.
      intros q q' v v' q'_eq.
      unfold dequeue, enqueue in *.
      generalize dependent q'.
      induction q;  intros q' q'_eq.
      inversion q'_eq.
      inversion q'_eq. subst.
      simpl. auto.
    Qed.
  End Types.
End ListQueue.
