(** Useful facts about sequential consistency. *)

Require Import init_wb.

Set Implicit Arguments.

Import L.
Import SC.
Import L_SC.
Import SCFacts.
Import L_Axioms.

Lemma heap_steps_nonempty : forall H H' e,
  heap_step H e H' ->
  exists a, e = Some a.
Proof.
  intros H H' e steps.
  inversion steps; eauto.
Qed.

(* A simple tactic we use in the asynchrony proof to show that two locations
   (or locks) are not equal. *)
Ltac different_locations r1 r2 not_conflict :=
  assert (r1 <> r2);
    [ intro; subst r2; apply not_conflict; simpl; auto; fail | idtac ].

Hint Constructors actions_conflict.

Lemma heap_asynchrony : forall H H1 H2 e1 e2,
  heap_step H (Some e1) H1 ->
  heap_step H1 (Some e2) H2 ->
  ~ effects_conflict e1 e2 ->
  exists H1',
    heap_step H (Some e2) H1'
    /\ heap_step H1' (Some e1) H2.
Proof with eauto with stores locks sequential_consistency.
  intros H H1 H2 e1 e2 steps1 steps2 not_conflict.

  destruct H as [ S Lambda ].
  destruct H1 as [ S1 Lambda1 ].
  destruct H2 as [ S2 Lambda2 ].
  destruct e1 as [ t1 a1 ].
  destruct e2 as [ t2 a2 ].

  assert (~ actions_conflict a1 a2) as not_conflict_actions.
  intro conflict. apply not_conflict.
  apply effects_conflict_actions. auto.

  assert (t1 <> t2) as concurrent.
  intro. subst. apply not_conflict. apply effects_conflict_program_order.

  induction a1 as [
  (* pure *)
    | r1 v1 (* alloc *)
    | r1 v1 (* read *)
    | r1 v1 (* write *)
    | l1 (* acquire *)
    | l1 (* release *)
    | t3 (* spawn *)
  ];
  induction a2 as [
  (* pure *)
    | r2 v2 (* alloc *)
    | r2 v2 (* read *)
    | r2 v2 (* write *)
    | l2 (* acquire *)
    | l2 (* release *)
    | t4 (* spawn *)
  ];
  inversion steps1; subst;
    inversion steps2; subst;
      try different_locations r1 r2 not_conflict_actions;
        try different_locations l1 l2 not_conflict_actions;
          esplit; split...
  rewrite Store.eq_update_update...
  rewrite Store.eq_update_update...
  rewrite Store.eq_update_update...
  rewrite Store.eq_update_update...
  rewrite LockMap.eq_acquire_acquire...
  rewrite LockMap.eq_release_acquire...
  rewrite LockMap.eq_acquire_release...
  rewrite LockMap.eq_release_release...
Qed.

Hint Constructors system_step.

Lemma step_not_empty : forall H P H' P' e,
  system_step H P e H' P' ->
  exists t, exists a, e = Some (t, a).
Proof.
  intros H P H' P' e steps.
  inversion steps as [ | ? ? ? heap_steps | ]; eauto.
  inversion heap_steps.
Qed.

Lemma step_many_sequential_trace : forall H P H' P' tr,
  system_step_many H P H' P' tr ->
  sequential_trace tr.
Proof.
  intros H P H' P' tr steps_many.
  induction steps_many as [ | H P H'' P'' H' P' tr e steps_many ? steps ].
  apply sequential_trace_nil.
  pose proof step_not_empty steps as [ ? [ ? ? ] ]. subst.
  apply sequential_trace_cons. auto.
Qed.

Lemma program_conflict_effects_conflict : forall e1 e2,
  ~ effects_conflict e1 e2 ->
  ~ program_conflict e1 e2.
Proof.
  intros e1 e2 not_conflict conflict.
  apply not_conflict.
  inversion conflict.
  apply effects_conflict_program_order.
  apply effects_conflict_spawn_left.
  apply effects_conflict_spawn_right.
Qed.

Lemma asynchrony : forall H P H1 P1 H2 P2 e1 e2,
  system_step H P (Some e1) H1 P1 ->
  system_step H1 P1 (Some e2) H2 P2 ->
  ~ effects_conflict e1 e2 ->
  exists H1', exists P1',
    system_step H P (Some e2) H1' P1'
    /\ system_step H1' P1' (Some e1) H2 P2.
Proof.
  intros H P H1 P1 H2 P2 e1 e2 steps1 steps2 not_conflict.
  forwards not_program_conflict : program_conflict_effects_conflict; eauto.
  inversion steps1 as [
    ? ? ? ? ? ? program_steps1 heap_steps1
    | ? ? ? heap_steps1
    | ? ? ? ? program_steps1
  ]; subst;
  inversion steps2 as [
    ? ? ? ? ? ? program_steps2 heap_steps2
    | ? ? ? heap_steps2
    | ? ? ? ? program_steps2
  ]; subst;
  try pose proof heap_asynchrony heap_steps1 heap_steps2 not_conflict
    as [ ? [ ? ? ] ];
      try pose proof program_asynchrony program_steps1 program_steps2
        not_program_conflict as [ ? [ ? ? ] ];
          eauto 6.
Qed.

Lemma computational_equivalence : forall H P H' P' tr tr',
  system_step_many H P H' P' tr ->
  sequential_trace tr' ->
  equivalence (erase_trace tr) (erase_trace tr') ->
  system_step_many H P H' P' tr'.
Proof.
  intros H P H' P' tr_ tr'_ steps_many seq_tr' equiv.
  remember (erase_trace tr_) as tr.
  remember (erase_trace tr'_) as tr'.

    (* Proof by induction on the derivation of equivalence. *)
  generalize dependent tr'_.
  generalize dependent tr_.
  generalize dependent P'.
  generalize dependent H'.
  generalize dependent P.
  generalize dependent H.
  induction equiv as [
    [ t1 a1 ] [ t2 a2 ] not_conflict12
    | tr1 tr2 tr1' tr2' equiv1 ? equiv2
    | tr
    | tr1 tr2 tr3 equiv1 ? equiv2
  ]; intros H P H' P' tr_ steps_many Heqtr tr'_ seq_tr' Heqtr';
  pose proof step_many_sequential_trace steps_many as seq_tr;
    pose proof sequential_trace_erase seq_tr as tr_eq;
      rewrite <- Heqtr in tr_eq; simpl in tr_eq; subst tr_;
        pose proof sequential_trace_erase seq_tr' as tr'_eq;
          rewrite <- Heqtr' in tr'_eq; simpl in tr'_eq; subst tr'_.

  Focus.
  (* The two traces consist of two non-conflicting, concurrent actions, in
     either order. *)
  inversion steps_many as [ | ? ? H2 P2 ? ? ? ? steps_many2 steps2 ].
  subst.
  inversion steps_many2 as [ | ? ? H1 P1 ? ? ? ? steps_many3 steps1 ].
  subst.
  inversion steps_many3. subst.
  assert (~ effects_conflict (t2, a2) (t1, a1)) as not_conflict21.
  intro conflict21. apply not_conflict12.
  apply effects_conflict_sym. auto.
  pose proof asynchrony steps1 steps2 not_conflict21
    as [ H1' [ P1' [ steps2' steps1' ] ] ].
  eapply system_step_many_trans; eauto.
  Unfocus.

  Focus.
  (* Inductive case #1: equivalence via appending two equivalent traces. *)
  rewrite app_unerase in *.
  pose proof sequential_trace_append_inversion _ _ seq_tr' as [ ? ? ].
  pose proof system_step_many_append_inversion _ _ steps_many
    as [ H'' [ P'' [ steps_tr2 steps_tr1 ] ] ].
  apply system_step_many_append with (H2 := H'') (P2 := P'');
    eauto using erase_unerase.
  Unfocus.

  Focus.
  (* The reflexivity case. *)
  auto.
  Unfocus.

  Focus.
  (* The transitivity case. *)
  apply IHequiv2 with (tr_ := unerase_trace tr2);
    auto using erase_unerase, sequential_trace_unerase.
  apply IHequiv1 with (tr_ := unerase_trace tr1);
    auto using erase_unerase, sequential_trace_unerase.
  Unfocus.
Qed.
