Require Import init_wb.

Inductive map_heap : SC.heap -> WB.heap -> Prop :=
| map_heap_def : forall S1 L S2 B2,
  (forall r,
    SC.Store.fresh S1 r ->
    WB.Store.fresh S2 r) ->
  (forall r v,
    SC.Store.get S1 r v ->
    WB.Store.get S2 r v) ->
  WB.Buffer.is_empty B2 ->
  map_heap (S1, L) (S2, B2, L).

Ltac inversion_map_heap MapHeap :=
  inversion MapHeap as [ ? ? ? ? Fresh Read BufferEmpty ]; subst.

Lemma map_heap_empty : map_heap SC.empty_heap WB.empty_heap.
Proof.
  apply map_heap_def.
  intros r Fresh.
  apply WB.Store.fresh_empty.
  intros r v Read.
  apply SC.Store.not_fresh_after_get in Read.
  contradiction Read.
  apply SC.Store.fresh_empty.
  apply WB.Buffer.empty_is_empty.
Qed.

Lemma map_heap_alloc : forall H1 H2 t r v H1',
  map_heap H1 H2 ->
  SC.heap_step H1 (Some (t, alloc r v)) H1' ->
  exists H2',
    WB.heap_step H2 (Some (t, alloc r v)) H2'
    /\ map_heap H1' H2'.
Proof with eauto.
  intros [ S1 L1 ] [ [ S2 B2 ] L2 ] t r v [ S1' L1' ] MapHeap Step.
  inversion_map_heap MapHeap.
  inversion Step; subst.
  esplit. split.
  apply WB.step_alloc...

  apply map_heap_def...

  intros r' Fresh1'.
  destruct (r == r').
  subst.
  forwards NotFresh1' : SC.Store.not_fresh_after_update.
  contradiction NotFresh1'...
  apply WB.Store.fresh_after_update...
  apply Fresh.
  eapply SC.Store.fresh_before_update...

  intros r' v' Get1'.
  destruct (r == r').
  subst.
  forwards : SC.Store.get_unique.
  apply Get1'.
  apply SC.Store.get_update_same.
  subst.
  apply WB.Store.get_update_same.
  apply WB.Store.get_after_update...
  apply Read.
  eapply SC.Store.get_before_update...
Qed.

Lemma map_heap_read : forall H1 H2 t r v,
  map_heap H1 H2 ->
  SC.heap_step H1 (Some (t, read r v)) H1 ->
  WB.heap_step H2 (Some (t, read r v)) H2.
Proof with eauto.
  intros [ S1 L1 ] [ [ S2 B2 ] L2 ] t r v MapHeap Step.
  inversion_map_heap MapHeap.
  inversion Step; subst.
  apply WB.step_read...
Qed.

Lemma map_heap_write : forall H1 H2 t r v H1',
  map_heap H1 H2 ->
  SC.heap_step H1 (Some (t, write r v)) H1' ->
  exists H2', exists H2'',
    WB.heap_step H2 (Some (t, write r v)) H2'
    /\ WB.heap_step H2' None H2''
    /\ map_heap H1' H2''.
Proof with eauto.
  intros [ S1 L1 ] [ [ S2 B2 ] L2 ] t r v [ S1' L1' ] MapHeap Step.
  inversion_map_heap MapHeap.
  inversion Step; subst.
  forwards [ ? [ ? ? ] ] : WB.Buffer.dequeue_enqueue_empty...
  esplit. esplit. split.
  apply WB.step_write.
  split.
  eapply WB.step_commit...

  apply map_heap_def...

  intros r' Fresh1'.
  destruct (r == r').
  subst.
  forwards NotFresh1' : SC.Store.not_fresh_after_update.
  contradiction NotFresh1'...
  apply WB.Store.fresh_after_update...
  apply Fresh.
  eapply SC.Store.fresh_before_update...

  intros r' v' Read1'.
  destruct (r == r').
  subst.
  forwards : SC.Store.get_unique.
  apply Read1'.
  apply SC.Store.get_update_same.
  subst.
  apply WB.Store.get_update_same.
  apply WB.Store.get_after_update...
  apply Read.
  eapply SC.Store.get_before_update...
Qed.

Lemma map_heap_acquire : forall H1 H2 t l H1',
  map_heap H1 H2 ->
  SC.heap_step H1 (Some (t, acquire _ l)) H1' ->
  exists H2',
    WB.heap_step H2 (Some (t, acquire _ l)) H2'
    /\ map_heap H1' H2'.
Proof with eauto.
  intros [ S1 L1 ] [ [ S2 B2 ] L2 ] t l [ S1' L1' ] MapHeap Step.
  inversion_map_heap MapHeap.
  inversion Step; subst.
  esplit. split.
  apply WB.step_acquire...
  apply map_heap_def...
Qed.

Lemma map_heap_release : forall H1 H2 t l H1',
  map_heap H1 H2 ->
  SC.heap_step H1 (Some (t, release _ l)) H1' ->
  exists H2',
    WB.heap_step H2 (Some (t, release _ l)) H2'
    /\ map_heap H1' H2'.
Proof with eauto.
  intros [ S1 L1 ] [ [ S2 B2 ] L2 ] t l [ S1' L1' ] MapHeap Step.
  inversion_map_heap MapHeap.
  inversion Step; subst.
  esplit. split.
  apply WB.step_release...
  unfold WB.Buffer.is_empty_tid, WB.Buffer.is_empty in *...
  apply map_heap_def...
Qed.

Lemma map_heap_spawn : forall H1 H2 t t',
  map_heap H1 H2 ->
  SC.heap_step H1 (Some (t, spawn _ t')) H1 ->
  WB.heap_step H2 (Some (t, spawn _ t')) H2.
Proof with eauto.
  intros H1 H2 t t' MapHeap Step.
  inversion_map_heap MapHeap.
  inversion Step; subst.
  apply WB.step_spawn.
  unfold WB.Buffer.is_empty_tid, WB.Buffer.is_empty in *...
Qed.

Lemma sc_to_wb_step_many : forall H1 H2 P H1' P' tr1,
  map_heap H1 H2 ->
  L_SC.system_step_many H1 P H1' P' tr1 ->
  exists H2', exists tr2,
    L_WB.system_step_many H2 P H2' P' tr2
    /\ map_heap H1' H2'.
Proof with eauto.
  intros H1 H2 P H1' P' tr1 MapHeap StepMany.
  induction StepMany as [ | ? ? ? ? ? ? ? ? ? ? Step ].

  esplit. esplit. split...

  specializes IHStepMany (>>> __)...
  destruct IHStepMany as [ ? [ ? [ ? ? ] ] ].
  inversion Step as [ ? ? ? ? ? ? ? HeapStep | ? ? ? HeapStep | ]; subst.
  inversion HeapStep; subst.

  forwards [ ? [ ? ? ] ] : map_heap_alloc...
  esplit. esplit. split...
  eapply L_WB.system_step_many_trans...
  apply L_WB.system_step_mutual...

  forwards : map_heap_read...
  esplit. esplit. split...
  eapply L_WB.system_step_many_trans...
  apply L_WB.system_step_mutual...

  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : map_heap_write...
  esplit. esplit. split...
  eapply L_WB.system_step_many_trans.
  eapply L_WB.system_step_many_trans...
  apply L_WB.system_step_mutual...
  apply L_WB.system_step_heap...
  
  forwards [ ? [ ? ? ] ] : map_heap_acquire...
  esplit. esplit. split...
  eapply L_WB.system_step_many_trans...
  apply L_WB.system_step_mutual...

  forwards [ ? [ ? ? ] ] : map_heap_release...
  esplit. esplit. split...
  eapply L_WB.system_step_many_trans...
  apply L_WB.system_step_mutual...

  forwards : map_heap_spawn...
  esplit. esplit. split...
  eapply L_WB.system_step_many_trans...
  apply L_WB.system_step_mutual...

  inversion HeapStep.

  esplit. esplit. split...
  eapply L_WB.system_step_many_trans...
  apply L_WB.system_step_program...
Qed.

Theorem sc_to_wb : forall P0 H1 P tr,
  L_SC.system_step_many SC.empty_heap P0 H1 P tr ->
  exists H2, exists tr',
    L_WB.system_step_many WB.empty_heap P0 H2 P tr'.
Proof with eauto.
  intros P0 H1' P tr StepMany.
  forwards MapHeap : map_heap_empty.
  forwards [ ? [ ? [ ? ? ] ] ] : sc_to_wb_step_many...
Qed.
