Require Import Metatheory.

Require Import init_wb.

(** This module guarantees that sequential consistency is type-preserving (for
   type systems that satisfy certain reasonable properties. *)
Module SC_Type_Preservation <: MEMORY_MODEL_TYPE_PRESERVATION Typ.
  Module M := SC.

  Import Typ.
  Import SC.

  (** A heap (S, Lambda) is well-typed with respect to a heap typing Sigma if
     the domains of S and Sigma are equal, and all of the values stored in S
     have the type given by Sigma. *)
  Definition well_typed (Sigma : heap_typing) (H : heap) : Prop :=
    let (S, _) := H in
      (forall r, ~ Map.In r Sigma -> Store.fresh S r) /\
      (forall r T, Map.MapsTo r T Sigma ->
        exists v, Store.get S r v /\ typing_val Sigma v T).

  Hint Constructors typing_incoming.
  Hint Unfold well_typed.

  (** The empty heap is well-typed. *)
  Lemma well_typed_empty : well_typed empty_heap_typing empty_heap.
  Proof.
    split.
    intros. apply Store.fresh_empty.
    intros r T H.
    contradiction (Map.empty_1 H).
  Qed.

  (** A well-typed heap produces well-typed values for reads and fresh locations
     for allocations. *)
  Lemma typing_step : forall Sigma H t a H',
    well_typed Sigma H ->
    heap_step H (Some (t, a)) H' ->
    typing_incoming Sigma a.
  Proof.
    introv Typ Step.
    destruct H as [ S L ].
    unfold well_typed in Typ.
    destruct Typ as [ Dom Typ ].
    inversions* Step.

    (* alloc *)
    apply typing_incoming_alloc.
      intros [ T Maps ].
      specializes* Typ (>>>Hyps __).
      destruct Typ as [ v' [ Get Typ ] ].
      apply Store.not_fresh_after_get in Get.
      tauto.

    (* read *)
    remember (Map.find r Sigma) as maybe_T.
      symmetry in Heqmaybe_T.
      destruct maybe_T as [ T | ].
      apply Map.find_2 in Heqmaybe_T.
      apply* typing_incoming_read.
      specializes* Typ (>>> Hyps __).
      destruct Typ as [ v' [ Get Typ ] ].
      assert (v = v'). eapply Store.get_unique; eauto. subst. auto.
      forwards fr : Dom r.
      intros [ T Maps ].
      apply Map.find_1 in Maps.
      rewrite Heqmaybe_T in Maps. inversion Maps.
      apply Store.not_fresh_after_get in H4.
      tauto.
  Qed.

  Module TypeFacts := Type_Model_Facts Typ.
  Module VarMapFacts := VarMapFacts Map.

  Hint Unfold Var_as_OT.eq.

  (** Type preservation for mutual steps. *)
  Lemma preservation : forall Sigma H t a H' Sigma',
    well_typed Sigma H ->
    heap_step H (Some (t, a)) H' ->
    typing_outgoing Sigma a Sigma' ->
    well_typed Sigma' H'.
  Proof.
    introv Typ Step Out.
    forwards Inc : typing_step; eauto.
    forwards Ext : TypeFacts.typing_extends; eauto.
    destruct H as [ S Lambda ].
    unfold well_typed in Typ.
    destruct Typ as [ Dom Typ ].
    inversions* Step; inversions* Out; inversions* Inc; split.

    (* alloc - doms eq *)
    intros r' not_in.
    apply* Store.fresh_after_update.
    apply* Dom. intro in_sigma.
    unfold Map.In in *. destruct in_sigma.
    apply not_in. destruct (r == r').
    esplit. apply* Map.add_1.
    esplit. apply* Map.add_2.
    intro. subst.
    apply not_in. unfold Map.In.
    esplit. apply* Map.add_1.

    (* alloc - well-typed *)
    intros r' T' Maps.
    destruct (r == r'). subst.
      assert (T = T'). eapply VarMapFacts.MapsTo_unique.
        apply Map.add_1. auto. eauto. subst.
      esplit. split.
        apply Store.get_update_same.
        apply* typing_val_stability.
    forwards Typ' : Typ (>>> __ __ __).
      eapply Map.add_3; eauto.
      destruct Typ' as [ v' [ Get Typ' ] ].
      exists v'. split.
      apply* Store.get_after_update.
      apply* typing_val_stability.

    (* write - doms eq *)
    intros r' not_in.
    apply Store.fresh_after_update.
    apply Dom. auto.
    intro. subst.
    unfold Map.In in *.
    apply* not_in.

    (* write - well-typed *)
    intros r' T' Maps.
    destruct (r == r'). subst.
      exists v. split.
      apply Store.get_update_same.
      assert (T = T').
      eapply VarMapFacts.MapsTo_unique; eauto.
      subst. auto.
    specializes* Typ (>>> r' T' __).
      destruct Typ as [ ? [ ? ? ] ].
      esplit. split*.
      apply* Store.get_after_update.
  Qed.

  (** Type preservation for mutual steps. *)
  Lemma preservation_empty : forall Sigma H H',
    well_typed Sigma H ->
    heap_step H None H' ->
    well_typed Sigma H'.
  Proof.
    introv Typ Step.
    inversion Step.
  Qed.
End SC_Type_Preservation.
