Require Import Arith.
Require Import List.

Require Import actions.
Require Import indexes.
Require Import list_facts.

Set Implicit Arguments.

Open Scope nat_scope.

Section Subtraces.
  Variable val : Set.

  Inductive subtrace (tr : erased_trace val) (begin len : nat) :
    erased_trace val -> Prop :=
  | subtrace_def : forall tr1 tr2 tr3,
    tr = tr1 ++ tr2 ++ tr3 ->
    length tr1 = begin ->
    length tr2 = len ->
    subtrace tr begin len tr2.

  Lemma subtrace_unique : forall tr begin len tr1 tr2,
    subtrace tr begin len tr1 ->
    subtrace tr begin len tr2 ->
    tr1 = tr2.
  Proof.
    intros tr begin len tr' tr'' sub' sub''.
    inversion sub' as [ tr1' tr2' tr3' tr_eq' length_tr1' length_tr2' tr'_eq ].
    inversion sub''
      as [ tr1'' tr2'' tr3'' tr_eq'' length_tr1'' length_tr2'' tr''_eq ].
    subst begin len tr tr' tr''.
    symmetry in length_tr1''.
    pose proof append_begin_unique _ _ _ _ tr_eq'' length_tr1'' as [ ? tr23_eq ].
    subst tr1''.
    symmetry in length_tr2''.
    pose proof append_begin_unique _ _ _ _ tr23_eq length_tr2'' as [  ].
    auto.
  Qed.

  Lemma subtrace_inversion_length : forall tr begin len tr1,
    subtrace tr begin len tr1 ->
    length tr1 = len.
  Proof.
    intros tr begin len tr1 subtrace_tr1.
    inversion subtrace_tr1. auto.
  Qed.

  Lemma subtrace_inversion_begin : forall tr len tr1,
    subtrace tr 0 len tr1 ->
    exists tr2, tr = tr1 ++ tr2.
  Proof.
    intros tr len tr1 subtrace_tr1.
    inversion subtrace_tr1 as [ tr0 ].
    assert (tr0 = nil). rewrite length_nil. auto. subst tr0.
    simpl in *. eauto.
  Qed.

  (* Useful in the transposition lemma. Not a general result or anything. *)
  Lemma subtrace_inversion_2 : forall tr tr1 t1 a1 t2 a2 tr2 j len_tr2,
    subtrace tr 0 j tr1 ->
    index tr j (t1, a1) ->
    index tr (j + 1) (t2, a2) ->
    length tr = len_tr2 + j + 2 ->
    subtrace tr (j + 2) len_tr2 tr2 ->
    tr = tr1 ++ ((t1, a1) :: nil) ++ ((t2, a2) :: nil) ++ tr2.
  Proof.
    intros tr tr1 t1 a1 t2 a2 tr2 j len_jr2 subtrace_tr1 idx1 idx2 length_tr
      subtrace_tr2.

    (* tr1 is at the beginning *)
    inversion subtrace_tr1 as [ tr0 tr1_ tr3 tr_eq length_tr0 length_tr1 ].
    subst tr1_. rewrite <- length_nil in length_tr0. subst tr0.
    simpl in tr_eq.

    (* Then a1 *)
    assert (index tr3 0 (t1, a1)) as idx1_tr3.
    eapply index_append_inversion_2.
    rewrite plus_0_l. subst tr j. eauto.
    inversion idx1_tr3 as [ tr4 [ t1_ a1_ ] tr3_eq | ]. subst t1_ a1_ tr3.
    rewrite cons_nil_app in tr_eq.

    (* Then a2 *)
    assert (index tr4 0 (t2, a2)) as idx2_tr4.
    eapply index_append_inversion_2 with (l1 := tr1 ++ ((t1, a1) :: nil)).
    rewrite <- app_ass in tr_eq. rewrite tr_eq in idx2.
    rewrite app_length. simpl. subst j. auto.
    inversion idx2_tr4 as [ tr5 [ t2_ a2_ ] tr4_eq | ]. subst t2_ a2_ tr4.

    (* Then tr2 *)
    inversion subtrace_tr2 as [ tr6 tr2_ tr0 tr_eq' length_tr6 length_tr2 ].
    subst tr2_.
    assert (length tr0 = 0) as length_tr0.
    rewrite tr_eq' in length_tr. repeat rewrite app_length in length_tr.
    rewrite length_tr6 in length_tr. rewrite length_tr2 in length_tr.
    omega. rewrite <- length_nil in length_tr0. subst tr0.
    rewrite <- app_nil_end in tr_eq'.
    assert (tr6 = tr1 ++ ((t1, a1) :: nil) ++ ((t2, a2) :: nil) /\ tr2 = tr5)
      as [ ? ? ].
    apply append_begin_unique with (l1 := tr6) (l2 := tr2).
    subst tr. repeat rewrite app_ass. simpl in *. symmetry. auto.
    rewrite length_tr6. repeat rewrite app_length. simpl. subst j. auto.
    subst. simpl. auto.
  Qed.

  Lemma subtrace_plus : forall tr begin len tr1 a,
    subtrace tr begin len tr1 ->
    index tr (begin + len) a ->
    subtrace tr begin (len + 1) (tr1 ++ (a :: nil)).
  Proof.
    intros tr begin len tr1 a subtrace_tr1 idx.
    inversion subtrace_tr1 as [ tr0 tr1' tr2 tr_eq length_tr0 length_tr1 ].
    subst tr1' begin len tr.
    rewrite plus_comm in idx.
    pose proof index_append_inversion_2 _ _ _ idx as idx'.
    rewrite <- (plus_0_l (length tr1)) in idx'.
    pose proof index_append_inversion_2 _ _ _ idx' as idx''.
    inversion idx'' as [ tr3 [ t' a' ] | ]. subst.
    eapply subtrace_def with (tr1 := tr0).
    rewrite app_ass. simpl. eauto. auto.
    rewrite app_length. simpl. auto.
  Qed.

  Lemma subtrace_minus : forall tr begin len tr1 a,
    subtrace tr begin (len + 1) (tr1 ++ (a :: nil)) ->
    subtrace tr begin len tr1.
  Proof.
    intros tr begin len tr1 a sub.
    inversion sub. subst.
    eapply subtrace_def.
    rewrite app_ass. auto.
    auto. rewrite app_length in *. simpl in *.
    omega.
  Qed.

  Lemma subtrace_plus_inversion : forall tr begin len tr1 a,
    subtrace tr begin (len + 1) (tr1 ++ (a :: nil)) ->
    index tr (begin + len) a.
  Proof.
    intros tr begin len tr1 a subtrace_tr1.
    inversion subtrace_tr1 as [ tr0 tr1' tr2 tr_eq length_tr0 len_eq ].
    subst.
    rewrite plus_comm.
    apply index_append_2.
    rewrite app_ass.
    rewrite <- (plus_0_l len).
    rewrite app_length in len_eq. simpl in len_eq.
    assert (len = length tr1). omega. subst len.
    apply index_append_2.
    simpl. apply index_0.
  Qed.

  Lemma subtrace_extend_begin : forall tr begin len_tr1 len_tr2 tr2,
    subtrace tr (begin + len_tr1) (len_tr2) tr2 ->
    exists tr1, subtrace tr begin (len_tr1 + len_tr2) (tr1 ++ tr2).
  Proof.
    intros tr begin len_tr1 len_tr2 tr2 subtrace_tr2.
    inversion subtrace_tr2 as [ tr0 tr2' tr3 tr_eq length_tr0 length_tr2 ].
    subst tr2' len_tr2 tr.
    pose proof length_append_exists _ _ _ length_tr0
      as [ tr0' [ tr1 [ ? [ ? ? ] ] ] ].
    subst tr0. subst len_tr1.
    exists tr1. rewrite app_ass.
    eapply subtrace_def with (tr1 := tr0') (tr3 := tr3).
    rewrite app_ass. auto. auto. apply app_length.
  Qed.

  Lemma subtrace_extend_begin_inversion : forall tr begin len_tr1 len_tr2 tr1 tr2,
    subtrace tr begin (len_tr1 + len_tr2) (tr1 ++ tr2) ->
    len_tr1 = length tr1 ->
    subtrace tr (begin + len_tr1) (len_tr2) tr2.
  Proof.
    intros tr begin len1 len2 tr1 tr2 sub length_tr1.
    inversion sub as [ tr0 ? tr3 ]. subst.
    eapply subtrace_def with (tr1 := tr0 ++ tr1) (tr3 := tr3).
    rewrite app_ass. rewrite <- app_ass. auto.
    rewrite app_length. auto.
    rewrite app_length in *. omega.
  Qed.
End Subtraces.
