Require Import Metatheory.

Require Import models.
Require Import var_map.

Set Implicit Arguments.

(** A type model enriches a specific language model with a type system. *)
Module Type TYPE_MODEL.
  Declare Module L : LANGUAGE_MODEL.

  (** The types for the language. *)
  Parameter type : Set.

  (** A heap typing is a map from locations to types. *)
  Definition heap_typing := Map.t type.
  Definition empty_heap_typing := Map.empty type.

  (** Sigma' extends Sigma if it contains all of Sigma's mapping. *)
  Definition extends (Sigma Sigma' : heap_typing) :=
    forall r T, Map.MapsTo r T Sigma ->
      Map.MapsTo r T Sigma'.
  Hint Unfold extends.

  (** Typing judgments for values and program states. *)
  Parameter typing_val : heap_typing -> L.val -> type -> Prop.
  Parameter typing : heap_typing -> L.program_state -> Prop.

  (** Value typing holds for an extending heap typing. *)
  Axiom typing_val_stability : forall Sigma v T Sigma',
    typing_val Sigma v T ->
    extends Sigma Sigma' ->
    typing_val Sigma' v T.

  (** The heap must ensure that any values read are well-typed and that any
     locations allocated are fresh. *)
  Inductive typing_incoming : heap_typing -> action L.val -> Prop :=
  | typing_incoming_alloc : forall Sigma r v,
    ~ Map.In r Sigma ->
    typing_incoming Sigma (alloc r v)
  | typing_incoming_read : forall Sigma r v T,
    Map.MapsTo r T Sigma ->
    typing_val Sigma v T ->
    typing_incoming Sigma (read r v)

  | typing_incoming_write : forall Sigma r v,
    typing_incoming Sigma (write r v)
  | typing_incoming_acquire : forall Sigma l,
    typing_incoming Sigma (acquire _ l)
  | typing_incoming_release : forall Sigma l,
    typing_incoming Sigma (release _ l)
  | typing_incoming_spawn : forall Sigma id,
    typing_incoming Sigma (spawn _ id).

  (** The program must ensure that any values written or allocated are
     well-typed. *)
  Inductive typing_outgoing :
    heap_typing -> action L.val -> heap_typing -> Prop :=
  | typing_outgoing_alloc : forall Sigma r v T,
    typing_val Sigma v T ->
    typing_outgoing Sigma (alloc r v) (Map.add r T Sigma)
  | typing_outgoing_write : forall Sigma r v T,
    Map.MapsTo r T Sigma ->
    typing_val Sigma v T ->
    typing_outgoing Sigma (write r v) Sigma

  | typing_outgoing_read : forall Sigma r v,
    typing_outgoing Sigma (read r v) Sigma
  | typing_outgoing_acquire : forall Sigma l,
    typing_outgoing Sigma (acquire _ l) Sigma
  | typing_outgoing_release : forall Sigma l,
    typing_outgoing Sigma (release _ l) Sigma
  | typing_outgoing_spawn : forall Sigma id,
    typing_outgoing Sigma (spawn _ id) Sigma.
End TYPE_MODEL.

(** Facts about heap typings. *)
Module Type_Model_Facts (Typ : TYPE_MODEL).
  Import Typ.

  Lemma typing_extends : forall Sigma a Sigma',
    typing_incoming Sigma a ->
    typing_outgoing Sigma a Sigma' ->
    extends Sigma Sigma'.
  Proof.
    introv Inc Out.
    inductions a; inversions Out; inversions Inc; auto*.
    intros l' T' Maps. destruct (l == l').
    subst. unfold Map.In in *. false*.
    apply* Map.add_2.
  Qed.
End Type_Model_Facts.

(** This module guarantees that a given type system is type-preserving (assuming
   a type-preserving memory model). *)
Module Type TYPE_MODEL_PRESERVATION.
  Declare Module L : LANGUAGE_MODEL.
  Declare Module Typ : TYPE_MODEL with Module L := L.

  Import L.
  Import Typ.

  (** Progress does not hold (due to locking). Preservation only holds in
     combination with a type-preserving heap semantics. This axiom guarantees
     that the program has type preservation for mutual steps. *)
  Axiom preservation : forall Sigma P t a P',
    typing Sigma P ->
    program_step P (Some (t, a)) P' ->
    typing_incoming Sigma a ->
    exists Sigma',
      typing_outgoing Sigma a Sigma' /\
      typing Sigma' P'.

  (** Type preservation for pure steps. *)
  Axiom preservation_pure : forall Sigma P id P',
    typing Sigma P ->
    program_step P (Some (id, pure _)) P' ->
    typing Sigma P'.
End TYPE_MODEL_PRESERVATION.

(** This module guarantees that a specific memory model is type-preserving. *)
Module Type MEMORY_MODEL_TYPE_PRESERVATION (Typ : TYPE_MODEL).
  Declare Module M : MEMORY_MODEL Typ.L.

  Import Typ.
  Import M.

  (** This predicate on heaps guarantees that any reads will produce well-typed
     values. E.g., for sequential consistency, we say that the domains of Sigma
     and H must be equal, and that all values in the heap have the type given
     by Sigma. *)
  Parameter well_typed : heap_typing -> heap -> Prop.

  (** The empty heap is well-typed. *)
  Axiom well_typed_empty : well_typed empty_heap_typing empty_heap.

  (** A well-typed heap produces well-typed values for reads. *)
  Axiom typing_step : forall Sigma H t a H',
    well_typed Sigma H ->
    heap_step H (Some (t, a)) H' ->
    typing_incoming Sigma a.

  (** Type preservation for mutual steps. *)
  Axiom preservation : forall Sigma H t a H' Sigma',
    well_typed Sigma H ->
    heap_step H (Some (t, a)) H' ->
    typing_outgoing Sigma a Sigma' ->
    well_typed Sigma' H'.

  (** Type preservation for empty steps. *)
  Axiom preservation_empty : forall Sigma H H',
    well_typed Sigma H ->
    heap_step H None H' ->
    well_typed Sigma H'.
End MEMORY_MODEL_TYPE_PRESERVATION.
