Require Import Decidable.
Require Import FMaps.
Require Import Metatheory.

Require Import list_facts.

Set Implicit Arguments.

Import Var_as_OT.

Module Type VAR_MAP.
  Include Type S with Module E := Var_as_OT.

  Section Types.
    Variable elt : Type.

    (* We say that functionally equivalent maps are equal. *)
    Axiom extensionality : forall (M1 M2 : t elt),
      Equal M1 M2 -> M1 = M2.
  End Types.
End VAR_MAP.

Declare Module Map : VAR_MAP.

Module VarMapFacts (Import Map : VAR_MAP).
  Module MapFacts := FMapFacts.Facts Map.

  Section Maps.
    Variable val : Type.
    Definition t := t val.

    Lemma in_after_add : forall (M : t) k1 k2 v2,
      Map.In k1 M ->
      Map.In k1 (Map.add k2 v2 M).
    Proof.
      intros M k1 k2 v2 in_before.
      MapFacts.map_iff. auto.
    Qed.

    Lemma in_before_add : forall (M : t) k1 k2 v2,
      Map.In k1 (Map.add k2 v2 M) ->
      k1 <> k2 ->
      Map.In k1 M.
    Proof.
      intros M k1 k2 v2 in_after k1_neq.
      rewrite <- MapFacts.add_neq_in_iff; eauto.
    Qed.

    Lemma in_after_remove : forall (M : t) k1 k2,
      Map.In k1 M ->
      k1 <> k2 ->
      Map.In k1 (Map.remove k2 M).
    Proof.
      intros M k1 k2 in_before k1_neq.
      MapFacts.map_iff. auto.
    Qed.

    Lemma in_before_remove : forall (M : t) k1 k2,
      Map.In k1 (Map.remove k2 M) ->
      k1 <> k2 ->
      Map.In k1 M.
    Proof.
      intros M k1 k2 in_after k1_neq.
      rewrite <- MapFacts.remove_neq_in_iff; eauto.
    Qed.

    Lemma eq_add_add : forall (M : t) k1 k2 v1 v2,
      k1 <> k2 ->
      Map.Equal (Map.add k2 v2 (Map.add k1 v1 M))
      (Map.add k1 v1 (Map.add k2 v2 M)).
    Proof with auto.
      intros M k1 k2 v1 v2 k1_neq k3.
      destruct (eq_var_dec k1 k3). subst.
      rewrite MapFacts.add_neq_o...
      rewrite MapFacts.add_eq_o...
      rewrite MapFacts.add_eq_o...
      destruct (eq_var_dec k2 k3). subst.
      rewrite MapFacts.add_eq_o...
      rewrite MapFacts.add_neq_o...
      rewrite MapFacts.add_eq_o...
      rewrite MapFacts.add_neq_o...
      rewrite MapFacts.add_neq_o...
      rewrite MapFacts.add_neq_o...
      rewrite MapFacts.add_neq_o...
    Qed.

    Lemma eq_add_remove : forall (M : t) k1 k2 v2,
      k1 <> k2 ->
      Map.Equal (Map.add k2 v2 (Map.remove k1 M))
      (Map.remove k1 (Map.add k2 v2 M)).
    Proof with auto.
      intros M k1 k2 v2 k1_neq k3.
      destruct (eq_var_dec k1 k3). subst.
      rewrite MapFacts.add_neq_o...
      rewrite MapFacts.remove_eq_o...
      rewrite MapFacts.remove_eq_o...
      destruct (eq_var_dec k2 k3). subst.
      rewrite MapFacts.add_eq_o...
      rewrite MapFacts.remove_neq_o...
      rewrite MapFacts.add_eq_o...
      rewrite MapFacts.add_neq_o...
      rewrite MapFacts.remove_neq_o...
      rewrite MapFacts.remove_neq_o...
      rewrite MapFacts.add_neq_o...
    Qed.

    Lemma eq_remove_add : forall (M : t) k1 k2 v1,
      k1 <> k2 ->
      Map.Equal (Map.remove k2 (Map.add k1 v1 M))
      (Map.add k1 v1 (Map.remove k2 M)).
    Proof.
      auto using MapFacts.Equal_sym, eq_add_remove.
    Qed.

    Lemma eq_remove_remove : forall (M : t) k1 k2,
      k1 <> k2 ->
      Map.Equal (Map.remove k1 (Map.remove k2 M))
      (Map.remove k2 (Map.remove k1 M)).
    Proof with auto.
      intros M k1 k2 k1_neq k3.
      destruct (eq_var_dec k1 k3); subst.
      rewrite MapFacts.remove_eq_o...
      rewrite MapFacts.remove_neq_o...
      rewrite MapFacts.remove_eq_o...
      destruct (eq_var_dec k2 k3). subst.
      rewrite MapFacts.remove_neq_o...
      rewrite MapFacts.remove_eq_o...
      rewrite MapFacts.remove_eq_o...
      rewrite MapFacts.remove_neq_o...
      rewrite MapFacts.remove_neq_o...
      rewrite MapFacts.remove_neq_o...
      rewrite MapFacts.remove_neq_o...
    Qed.

    Lemma eq_refl : forall (M : t),
      Map.Equal M M.
    Proof.
      apply MapFacts.Equal_refl.
    Qed.

    Lemma In_elements : forall M x e,
      InA (@eq_key_elt val) (x, e) (elements M) ->
      List.In (x, e) (elements M).
    Proof.
      intros M x e in1.
      induction in1 as [ [ x' e' ] elts' eq_x_e | ].
      unfold eq_key_elt in eq_x_e. simpl in eq_x_e.
      unfold eq in eq_x_e. destruct eq_x_e. subst.
      simpl. auto.
      simpl. right. auto.
    Qed.

    Lemma InA_elements : forall M x e,
      List.In (x, e) (elements M) ->
      InA (@eq_key_elt val) (x, e) (elements M).
    Proof.
      intros M x e in1.
      induction (elements M) as [ | [ x' e' ] LL ].
      inversion in1.
      destruct in1 as [ head | rest ].
      inversion head; subst.
      apply InA_cons_hd.
      unfold eq_key_elt, eq. simpl. auto.
      apply InA_cons_tl. apply IHLL. auto.
    Qed.

    Variable P : val -> Prop.
    Variable P_dec : forall v, decidable (P v).

    Lemma not_for_all_exists_not : forall M,
      ~ (forall a v, MapsTo a v M -> P v) ->
      exists a, exists v, MapsTo a v M /\ ~ P v.
    Proof.
      intros M not_all.
      remember (elements M) as L.
      remember (fun (p : var * val) => P (snd p)) as P'.
      assert (exists p, List.In p L /\ ~ P' p) as [ [ a v ] [ in_L not_P' ] ].
      apply not_for_all_exists_not_list.
      intros [ a v ]. subst P'. simpl. auto.
      intro all_P'. apply not_all.
      intros a v maps_M.
      apply elements_1 in maps_M.
      apply In_elements in maps_M.
      subst L. pose proof all_P' _ maps_M.
      subst P'. simpl in *. auto.
      exists a. exists v. split.
      apply elements_2. apply InA_elements. subst L. auto.
      subst P'. simpl in *. auto.
    Qed.

    Lemma MapsTo_unique : forall (M : t) a v v',
      MapsTo a v M ->
      MapsTo a v' M ->
      v = v'.
    Proof.
      intros M a v v' maps1 maps2.
      apply Map.find_1 in maps1.
      apply Map.find_1 in maps2.
      rewrite maps1 in maps2.
      inversion maps2. auto.
    Qed.
  End Maps.

  Hint Resolve in_after_add in_before_add in_after_remove in_before_remove
    eq_add_add eq_add_remove eq_remove_add eq_remove_remove eq_refl :
    var_maps.
End VarMapFacts.
