Require Import buffers.
Require Import locks.
Require Import models.
Require Import stores.

Set Implicit Arguments.

Open Scope type_scope.

Module Write_Buffering (L : LANGUAGE_MODEL) <: MEMORY_MODEL L.
  Declare Module Store : STORE L.
  Declare Module Buffer : BUFFER L.

  (* A heap is a triple of a store, a write buffer and a lock map. *)
  Definition heap := (Store.store * Buffer.buffer * LockMap.lock_map).

  Definition empty_heap : heap :=
    (Store.empty, Buffer.empty, LockMap.empty).

  Inductive step : heap -> effect L.val -> heap -> Prop :=
  | step_alloc : forall S B Lambda t r v,
    Store.fresh S r ->
    step (S, B, Lambda) (Some (t, alloc r v)) (Store.update S r v, B, Lambda)
  | step_read : forall S B Lambda t r v,
    Buffer.is_empty_queue B t r ->
    Store.get S r v ->
    step (S, B, Lambda) (Some (t, read r v)) (S, B, Lambda)
  | step_write : forall S B Lambda t r v,
    step (S, B, Lambda) (Some (t, write r v))
    (S, Buffer.enqueue B t r v, Lambda)
  | step_commit : forall S B Lambda t r v B',
    Buffer.dequeue B t r = Some (v, B') ->
    step (S, B, Lambda) None (Store.update S r v, B', Lambda)
  | step_acquire : forall S B Lambda t l,
    LockMap.available Lambda l ->
    step (S, B, Lambda) (Some (t, acquire _ l))
    (S, B, LockMap.acquire Lambda l t)
  | step_release : forall S B Lambda t l,
    LockMap.held_by Lambda l t ->
    Buffer.is_empty_tid B t ->
    step (S, B, Lambda) (Some (t, release _ l))
    (S, B, LockMap.release Lambda l)
  | step_spawn : forall S B Lambda t1 t2,
    Buffer.is_empty_tid B t1 ->
    step (S, B, Lambda) (Some (t1, spawn _ t2)) (S, B, Lambda)
  | step_read2 : forall S B Lambda t r v,
    Buffer.top B t r = Some v ->
    step (S, B, Lambda) (Some (t, read r v)) (S, B, Lambda).

  Definition heap_step := step.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).

  Fact no_pure_steps : forall H H' t,
    ~ (heap_step H (Some (t, pure _)) H').
  Proof.
    intros H H' t steps. inversion steps.
  Qed.
End Write_Buffering.
